"""Group of subgroups for governance in Conway+ eras."""

import logging

from cardano_clusterlib import gov_action_group
from cardano_clusterlib import gov_committee_group
from cardano_clusterlib import gov_drep_group
from cardano_clusterlib import gov_vote_group
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class GovernanceGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj

        # Groups of commands
        self._action_group: gov_action_group.GovActionGroup | None = None
        self._committee_group: gov_committee_group.GovCommitteeGroup | None = None
        self._drep_group: gov_drep_group.GovDrepGroup | None = None
        self._vote_group: gov_vote_group.GovVoteGroup | None = None

    @property
    def action(self) -> gov_action_group.GovActionGroup:
        """Action group."""
        if not self._action_group:
            self._action_group = gov_action_group.GovActionGroup(
                clusterlib_obj=self._clusterlib_obj
            )
        return self._action_group

    @property
    def committee(self) -> gov_committee_group.GovCommitteeGroup:
        """Committee group."""
        if not self._committee_group:
            self._committee_group = gov_committee_group.GovCommitteeGroup(
                clusterlib_obj=self._clusterlib_obj
            )
        return self._committee_group

    @property
    def drep(self) -> gov_drep_group.GovDrepGroup:
        """Drep group."""
        if not self._drep_group:
            self._drep_group = gov_drep_group.GovDrepGroup(clusterlib_obj=self._clusterlib_obj)
        return self._drep_group

    @property
    def vote(self) -> gov_vote_group.GovVoteGroup:
        """Vote group."""
        if not self._vote_group:
            self._vote_group = gov_vote_group.GovVoteGroup(clusterlib_obj=self._clusterlib_obj)
        return self._vote_group

    def get_anchor_data_hash(
        self,
        text: str = "",
        file_binary: itp.FileType | None = None,
        file_text: itp.FileType | None = None,
    ) -> str:
        """Compute the hash of some anchor data.

        Args:
            text: A text to hash as UTF-8.
            file_binary: A path to the binary file to hash.
            file_text: A path to the text file to hash.

        Returns:
            str: A hash string.
        """
        if text:
            content_args = ["--text", text]
        elif file_binary:
            content_args = ["--file-binary", str(file_binary)]
        elif file_text:
            content_args = ["--file-text", str(file_text)]
        else:
            msg = "Either `text`, `file_binary` or `file_text` is needed."
            raise AssertionError(msg)

        out_hash = (
            self._clusterlib_obj.cli(
                ["cardano-cli", "hash", "anchor-data", *content_args], add_default_args=False
            )
            .stdout.rstrip()
            .decode("ascii")
        )

        return out_hash

    def get_script_hash(
        self,
        script_file: itp.FileType | None = None,
    ) -> str:
        """Compute the hash of a script.

        Args:
            script_file: A path to the text file to hash.

        Returns:
            str: A hash string.
        """
        # TODO: make it a top-level function to reflect `cardano-cli hash`
        out_hash = (
            self._clusterlib_obj.cli(
                ["cardano-cli", "hash", "script", "--script-file", str(script_file)],
                add_default_args=False,
            )
            .stdout.rstrip()
            .decode("ascii")
        )

        return out_hash

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: clusterlib_obj={id(self._clusterlib_obj)}>"



import enum
import typing as tp

DEFAULT_COIN: tp.Final[str] = "lovelace"
MAINNET_MAGIC: tp.Final[int] = 764824073
CONFIRM_BLOCKS_NUM: tp.Final[int] = 2

# The SUBCOMMAND_MARK is used to mark the beginning of a subcommand. It is used to differentiate
# between options and subcommands. That is needed for CLI coverage recording.
# For example, the command `cardano-cli query tx-mempool --cardano-mode info`
# has the following arguments:
#  ["query", "tx-mempool", "--cardano-mode", SUBCOMMAND_MARK, "info"]
SUBCOMMAND_MARK: tp.Final[str] = "SUBCOMMAND"


class CommandEras:
    CONWAY: tp.Final[str] = "conway"
    LATEST: tp.Final[str] = "latest"


class Eras(enum.Enum):
    BYRON = 1
    SHELLEY = 2
    ALLEGRA = 3
    MARY = 4
    ALONZO = 6
    BABBAGE = 8
    CONWAY = 9
    DEFAULT = CONWAY
    LATEST = CONWAY  # noqa: PIE796


class MultiSigTypeArgs:
    ALL: tp.Final[str] = "all"
    ANY: tp.Final[str] = "any"
    AT_LEAST: tp.Final[str] = "atLeast"


class MultiSlotTypeArgs:
    BEFORE: tp.Final[str] = "before"
    AFTER: tp.Final[str] = "after"


class ScriptTypes:
    SIMPLE_V1: tp.Final[str] = "simple_v1"
    SIMPLE_V2: tp.Final[str] = "simple_v2"
    PLUTUS_V1: tp.Final[str] = "plutus_v1"
    PLUTUS_V2: tp.Final[str] = "plutus_v2"
    PLUTUS_V3: tp.Final[str] = "plutus_v3"


class Votes(enum.Enum):
    YES = 1
    NO = 2
    ABSTAIN = 3



import dataclasses
import datetime
import pathlib as pl
import typing as tp

from cardano_clusterlib import consts
from cardano_clusterlib import types as itp


@dataclasses.dataclass(frozen=True)
class CLIOut:
    stdout: bytes
    stderr: bytes


@dataclasses.dataclass(frozen=True, order=True)
class KeyPair:
    vkey_file: pl.Path
    skey_file: pl.Path


@dataclasses.dataclass(frozen=True, order=True)
class ColdKeyPair:
    vkey_file: pl.Path
    skey_file: pl.Path
    counter_file: pl.Path


@dataclasses.dataclass(frozen=True, order=True)
class AddressRecord:
    address: str
    vkey_file: pl.Path
    skey_file: pl.Path


@dataclasses.dataclass(frozen=True, order=True)
class StakeAddrInfo:
    address: str
    delegation: str
    reward_account_balance: int
    registration_deposit: int
    vote_delegation: str

    def __bool__(self) -> bool:
        return bool(self.address)


@dataclasses.dataclass(frozen=True, order=True)
class UTXOData:
    utxo_hash: str
    utxo_ix: int
    amount: int
    address: str
    coin: str = consts.DEFAULT_COIN
    decoded_coin: str = ""
    datum_hash: str = ""
    inline_datum_hash: str = ""
    inline_datum: str | dict | None = None
    reference_script: dict | None = None


@dataclasses.dataclass(frozen=True, order=True)
class TxOut:
    address: str
    amount: int
    coin: str = consts.DEFAULT_COIN
    datum_hash: str = ""
    datum_hash_file: itp.FileType = ""
    datum_hash_cbor_file: itp.FileType = ""
    datum_hash_value: str = ""
    datum_embed_file: itp.FileType = ""
    datum_embed_cbor_file: itp.FileType = ""
    datum_embed_value: str = ""
    inline_datum_file: itp.FileType = ""
    inline_datum_cbor_file: itp.FileType = ""
    inline_datum_value: str = ""
    reference_script_file: itp.FileType = ""


# List of `TxOut`s, empty list, or empty tuple
OptionalTxOuts = list[TxOut] | tuple[()]
# List of `UTXOData`s, empty list, or empty tuple
OptionalUTXOData = list[UTXOData] | tuple[()]


@dataclasses.dataclass(frozen=True, order=True)
class ScriptTxIn:
    """Data structure for Tx inputs that are combined with scripts (simple or Plutus)."""

    txins: list[UTXOData]
    script_file: itp.FileType = ""
    reference_txin: UTXOData | None = None
    reference_type: str = ""
    # Values below needed only when working with Plutus
    collaterals: OptionalUTXOData = ()
    execution_units: tp.Optional[tuple[int, int]] = None
    datum_file: itp.FileType = ""
    datum_cbor_file: itp.FileType = ""
    datum_value: str = ""
    inline_datum_present: bool = False
    redeemer_file: itp.FileType = ""
    redeemer_cbor_file: itp.FileType = ""
    redeemer_value: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ScriptWithdrawal:
    """Data structure for withdrawals that are combined with Plutus scripts."""

    txout: TxOut
    script_file: itp.FileType = ""
    reference_txin: UTXOData | None = None
    reference_type: str = ""
    collaterals: OptionalUTXOData = ()
    execution_units: tp.Optional[tuple[int, int]] = None
    redeemer_file: itp.FileType = ""
    redeemer_cbor_file: itp.FileType = ""
    redeemer_value: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ComplexCert:
    """Data structure for certificates with optional data for Plutus scripts.

    If used for one certificate, it needs to be used for all the other certificates in a given
    transaction (instead of `TxFiles.certificate_files`). Otherwise, order of certificates
    cannot be guaranteed.
    """

    certificate_file: itp.FileType
    script_file: itp.FileType = ""
    reference_txin: UTXOData | None = None
    reference_type: str = ""
    collaterals: OptionalUTXOData = ()
    execution_units: tp.Optional[tuple[int, int]] = None
    redeemer_file: itp.FileType = ""
    redeemer_cbor_file: itp.FileType = ""
    redeemer_value: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ComplexProposal:
    """Data structure for proposal with optional data for Plutus scripts.

    If used for one proposal, it needs to be used for all the other proposals in a given
    transaction (instead of `TxFiles.proposal_files`). Otherwise, order of proposals
    cannot be guaranteed.
    """

    proposal_file: itp.FileType
    script_file: itp.FileType = ""
    collaterals: OptionalUTXOData = ()
    execution_units: tp.Optional[tuple[int, int]] = None
    redeemer_file: itp.FileType = ""
    redeemer_cbor_file: itp.FileType = ""
    redeemer_value: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ScriptVote:
    """Data structure for voting that are combined with scripts."""

    vote_file: itp.FileType = ""
    script_file: itp.FileType = ""
    # Values below needed only when working with Plutus
    collaterals: OptionalUTXOData = ()
    execution_units: tp.Optional[tuple[int, int]] = None
    redeemer_file: itp.FileType = ""
    redeemer_cbor_file: itp.FileType = ""
    redeemer_value: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class Mint:
    txouts: list[TxOut]
    script_file: itp.FileType = ""
    reference_txin: UTXOData | None = None
    reference_type: str = ""
    policyid: str = ""
    # Values below needed only when working with Plutus
    collaterals: OptionalUTXOData = ()
    execution_units: tp.Optional[tuple[int, int]] = None
    redeemer_file: itp.FileType = ""
    redeemer_cbor_file: itp.FileType = ""
    redeemer_value: str = ""


# List of `ScriptTxIn`s, empty list, or empty tuple
OptionalScriptTxIn = list[ScriptTxIn] | tuple[()]
# List of `ComplexCert`s, empty list, or empty tuple
OptionalScriptCerts = list[ComplexCert] | tuple[()]
# List of `ComplexProposal`s, empty list, or empty tuple
OptionalScriptProposals = list[ComplexProposal] | tuple[()]
# List of `ScriptWithdrawal`s, empty list, or empty tuple
OptionalScriptWithdrawals = list[ScriptWithdrawal] | tuple[()]
# List of `Mint`s, empty list, or empty tuple
OptionalMint = list[Mint] | tuple[()]
# List of `ScriptVote`s, empty list, or empty tuple
OptionalScriptVotes = list[ScriptVote] | tuple[()]


@dataclasses.dataclass(frozen=True, order=True)
class TxFiles:
    certificate_files: itp.OptionalFiles = ()
    proposal_files: itp.OptionalFiles = ()
    metadata_json_files: itp.OptionalFiles = ()
    metadata_cbor_files: itp.OptionalFiles = ()
    signing_key_files: itp.OptionalFiles = ()
    auxiliary_script_files: itp.OptionalFiles = ()
    vote_files: itp.OptionalFiles = ()
    metadata_json_detailed_schema: bool = False


@dataclasses.dataclass(frozen=True, order=True)
class PoolUser:
    payment: AddressRecord
    stake: AddressRecord


@dataclasses.dataclass(frozen=True, order=True)
class PoolData:
    pool_name: str
    pool_pledge: int
    pool_cost: int
    pool_margin: float
    pool_metadata_url: str = ""
    pool_metadata_hash: str = ""
    pool_relay_dns: str = ""
    pool_relay_ipv4: str = ""
    pool_relay_port: int = 0


@dataclasses.dataclass(frozen=True, order=True)
class TxRawOutput:
    txins: list[UTXOData]  # UTXOs used as inputs
    txouts: list[TxOut]  # Tx outputs
    txouts_count: int  # Final number of tx outputs after adding change address and joining outputs
    tx_files: TxFiles  # Files needed for transaction building (certificates, signing keys, etc.)
    out_file: pl.Path  # Output file path for the transaction body
    fee: int  # Tx fee
    build_args: list[str]  # Arguments that were passed to `cardano-cli transaction build*`
    era: str = ""  # Era used for the transaction
    script_txins: OptionalScriptTxIn = ()  # Tx inputs that are combined with scripts
    script_withdrawals: OptionalScriptWithdrawals = ()  # Withdrawals that are combined with scripts
    script_votes: OptionalScriptVotes = ()  # Votes that are combined with scripts
    complex_certs: OptionalScriptCerts = ()  # Certificates that are combined with scripts
    complex_proposals: OptionalScriptProposals = ()  # Proposals that are combined with scripts
    mint: OptionalMint = ()  # Minting data (Tx outputs, script, etc.)
    invalid_hereafter: int | None = None  # Validity interval upper bound
    invalid_before: int | None = None  # Validity interval lower bound
    current_treasury_value: int | None = None  # Current treasury value
    treasury_donation: int | None = None  # Amount of funds that will be donated to treasury
    withdrawals: OptionalTxOuts = ()  # All withdrawals (including those combined with scripts)
    change_address: str = ""  # Address for change
    return_collateral_txouts: OptionalTxOuts = ()  # Tx outputs for returning collateral
    total_collateral_amount: int | None = None  # Total collateral amount
    readonly_reference_txins: OptionalUTXOData = ()  # Tx inputs for plutus script context
    script_valid: bool = True  # Whether the plutus script is valid
    required_signers: itp.OptionalFiles = ()  # Signing keys required for the transaction
    # Hashes of signing keys that are required for the transaction
    required_signer_hashes: list[str] | tuple[()] = ()
    combined_reference_txins: OptionalUTXOData = ()  # All reference tx inputs


@dataclasses.dataclass(frozen=True, order=True)
class PoolCreationOutput:
    stake_pool_id: str
    vrf_key_pair: KeyPair
    cold_key_pair: ColdKeyPair
    pool_reg_cert_file: pl.Path
    pool_data: PoolData
    pool_owners: list[PoolUser]
    tx_raw_output: TxRawOutput
    kes_key_pair: KeyPair | None = None


@dataclasses.dataclass(frozen=True, order=True)
class GenesisKeys:
    genesis_utxo_vkey: pl.Path
    genesis_utxo_skey: pl.Path
    genesis_vkeys: list[pl.Path]
    delegate_skeys: list[pl.Path]


@dataclasses.dataclass(frozen=True, order=True)
class PoolParamsTop:
    pool_params: dict
    future_pool_params: dict
    retiring: int | None


@dataclasses.dataclass(frozen=True, order=True)
class AddressInfo:
    address: str
    era: str
    encoding: str
    type: str
    base16: str


@dataclasses.dataclass(frozen=True, order=True)
class Value:
    value: int
    coin: str


@dataclasses.dataclass(frozen=True, order=True)
class LeadershipSchedule:
    slot_no: int
    utc_time: datetime.datetime


@dataclasses.dataclass(frozen=True, order=True)
class DataForBuild:
    txins: list[UTXOData]
    txouts: list[TxOut]
    withdrawals: OptionalTxOuts
    script_withdrawals: OptionalScriptWithdrawals


@dataclasses.dataclass(frozen=True, order=True)
class CCMember:
    epoch: int
    cold_vkey: str = ""
    cold_vkey_file: itp.FileType = ""
    cold_vkey_hash: str = ""
    cold_skey: str = ""
    cold_skey_file: itp.FileType = ""
    cold_skey_hash: str = ""
    cold_script_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class VoteCC:
    action_txid: str
    action_ix: int
    vote: consts.Votes
    vote_file: pl.Path
    cc_hot_vkey: str = ""
    cc_hot_vkey_file: pl.Path | None = None
    cc_hot_key_hash: str = ""
    cc_hot_script_hash: str = ""
    anchor_url: str = ""
    anchor_data_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class VoteDrep:
    action_txid: str
    action_ix: int
    vote: consts.Votes
    vote_file: pl.Path
    drep_vkey: str = ""
    drep_vkey_file: pl.Path | None = None
    drep_key_hash: str = ""
    drep_script_hash: str = ""
    anchor_url: str = ""
    anchor_data_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class VoteSPO:
    action_txid: str
    action_ix: int
    vote: consts.Votes
    vote_file: pl.Path
    stake_pool_vkey: str = ""
    cold_vkey_file: pl.Path | None = None
    stake_pool_id: str = ""
    anchor_url: str = ""
    anchor_data_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ActionConstitution:
    action_file: pl.Path
    deposit_amt: int
    anchor_url: str
    anchor_data_hash: str
    constitution_url: str
    constitution_hash: str
    deposit_return_stake_vkey: str = ""
    deposit_return_stake_vkey_file: pl.Path | None = None
    deposit_return_stake_key_hash: str = ""
    prev_action_txid: str = ""
    prev_action_ix: int = -1


@dataclasses.dataclass(frozen=True, order=True)
class ActionInfo:
    action_file: pl.Path
    deposit_amt: int
    anchor_url: str
    anchor_data_hash: str
    deposit_return_stake_vkey: str = ""
    deposit_return_stake_vkey_file: pl.Path | None = None
    deposit_return_stake_key_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ActionNoConfidence:
    action_file: pl.Path
    deposit_amt: int
    anchor_url: str
    anchor_data_hash: str
    prev_action_txid: str
    prev_action_ix: int
    deposit_return_stake_vkey: str = ""
    deposit_return_stake_vkey_file: pl.Path | None = None
    deposit_return_stake_key_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ActionUpdateCommittee:
    action_file: pl.Path
    deposit_amt: int
    anchor_url: str
    anchor_data_hash: str
    threshold: str
    add_cc_members: list[CCMember] = dataclasses.field(default_factory=list)
    rem_cc_members: list[CCMember] = dataclasses.field(default_factory=list)
    prev_action_txid: str = ""
    prev_action_ix: int = -1
    deposit_return_stake_vkey: str = ""
    deposit_return_stake_vkey_file: pl.Path | None = None
    deposit_return_stake_key_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ActionPParamsUpdate:
    action_file: pl.Path
    deposit_amt: int
    anchor_url: str
    anchor_data_hash: str
    cli_args: itp.UnpackableSequence
    prev_action_txid: str = ""
    prev_action_ix: int = -1
    deposit_return_stake_vkey: str = ""
    deposit_return_stake_vkey_file: pl.Path | None = None
    deposit_return_stake_key_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ActionTreasuryWithdrawal:
    action_file: pl.Path
    transfer_amt: int
    deposit_amt: int
    anchor_url: str
    anchor_data_hash: str
    funds_receiving_stake_vkey: str = ""
    funds_receiving_stake_vkey_file: pl.Path | None = None
    funds_receiving_stake_key_hash: str = ""
    deposit_return_stake_vkey: str = ""
    deposit_return_stake_vkey_file: pl.Path | None = None
    deposit_return_stake_key_hash: str = ""


@dataclasses.dataclass(frozen=True, order=True)
class ActionHardfork:
    action_file: pl.Path
    deposit_amt: int
    anchor_url: str
    anchor_data_hash: str
    protocol_major_version: int
    protocol_minor_version: int
    prev_action_txid: str = ""
    prev_action_ix: int = -1
    deposit_return_stake_vkey: str = ""
    deposit_return_stake_vkey_file: pl.Path | None = None
    deposit_return_stake_key_hash: str = ""



"""Group of methods for working with key commands."""

import logging
import pathlib as pl

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import helpers
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class KeyGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj

    def gen_verification_key(
        self,
        key_name: str,
        signing_key_file: itp.FileType,
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Generate a verification file from a signing key.

        Args:
            key_name: A name of the key.
            signing_key_file: A path to signing key file.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the generated verification key file.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{key_name}.vkey"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                "key",
                "verification-key",
                "--signing-key-file",
                str(signing_key_file),
                "--verification-key-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def gen_non_extended_verification_key(
        self,
        key_name: str,
        extended_verification_key_file: itp.FileType,
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Generate a non-extended key from a verification key.

        Args:
            key_name: A name of the key.
            extended_verification_key_file: A path to the extended verification key file.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the generated non-extended verification key file.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{key_name}.vkey"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                "key",
                "non-extended-key",
                "--extended-verification-key-file",
                str(extended_verification_key_file),
                "--verification-key-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: clusterlib_obj={id(self._clusterlib_obj)}>"



"""Tools used by `ClusterLib` for constructing transactions."""

import base64
import contextlib
import dataclasses
import functools
import itertools
import logging

from cardano_clusterlib import consts
from cardano_clusterlib import exceptions
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


def _organize_tx_ins_outs_by_coin(
    tx_list: list[structs.UTXOData] | list[structs.TxOut] | tuple[()],
) -> dict[str, list]:
    """Organize transaction inputs or outputs by coin type."""
    db: dict[str, list] = {}
    for rec in tx_list:
        if rec.coin not in db:
            db[rec.coin] = []
        db[rec.coin].append(rec)
    return db


def _organize_utxos_by_id(
    tx_list: list[structs.UTXOData],
) -> dict[str, list[structs.UTXOData]]:
    """Organize UTxOs by ID (hash#ix)."""
    db: dict[str, list[structs.UTXOData]] = {}
    for rec in tx_list:
        utxo_id = f"{rec.utxo_hash}#{rec.utxo_ix}"
        if utxo_id not in db:
            db[utxo_id] = []
        db[utxo_id].append(rec)
    return db


def _organize_utxos_by_coin_and_id(
    tx_list: list[structs.UTXOData],
) -> dict[str, dict[str, int]]:
    """Organize UTxOs by coin and ID (hash#ix)."""
    db: dict[str, dict[str, int]] = {}
    for r in tx_list:
        utxo_id = f"{r.utxo_hash}#{r.utxo_ix}"
        db_rec = db.get(r.coin)
        if db_rec is None:
            db[r.coin] = {utxo_id: r.amount}
            continue
        db_rec[utxo_id] = r.amount
    return db


def _get_usable_utxos(
    address_utxos: list[structs.UTXOData], coins: set[str]
) -> list[structs.UTXOData]:
    """Get all UTxOs with no datum that contain any of the required coins (`coins`)."""
    txins_by_id = _organize_utxos_by_id(address_utxos)

    txins = []
    seen_ids = set()
    matching_with_datum = False
    for rec in address_utxos:
        utxo_id = f"{rec.utxo_hash}#{rec.utxo_ix}"
        if rec.coin in coins and utxo_id not in seen_ids:
            # Don't select UTxOs with datum
            if rec.datum_hash or rec.inline_datum_hash:
                matching_with_datum = True
                continue
            seen_ids.add(utxo_id)
            txins.extend(txins_by_id[utxo_id])

    if not txins and matching_with_datum:
        msg = "The only matching UTxOs have datum."
        raise exceptions.CLIError(msg)

    return txins


def _pick_coins_from_already_selected_utxos(
    coin_txins: dict[str, int],
    already_selected_utxos: set[str],
    target_amount: int,
    target_with_change: int,
) -> tuple[set[str], int, bool]:
    """Pick coins from already selected UTxOs if they have the desired coin.

    Args:
        coin_txins (dict): A dictionary of coin UTxOs.
        already_selected_utxos (set): A set of already selected UTxOs (for different coins).
        target_amount (int): The desired amount.
        target_with_change (int): The desired amount with minimal change.

    Returns:
        tuple: A tuple with selected UTxO IDs, accumulated amount and a bool indicating if the
            desired amount was met.
    """
    picked_utxos: set[str] = set()
    accumulated_amount = 0

    # See if the coin exists in UTxOs that were already selected
    for utxo_id in already_selected_utxos:
        utxo_amount = coin_txins.get(utxo_id)
        if utxo_amount is None:
            continue
        accumulated_amount += utxo_amount

        # If we were able to collect exact amount, no change is needed
        if accumulated_amount == target_amount:
            break
        # Make sure the change is higher than `_min_change_value`
        if accumulated_amount >= target_with_change:
            break
    else:
        return picked_utxos, accumulated_amount, False

    return picked_utxos, accumulated_amount, True


def _pick_utxos_with_defragmentation(
    utxos: list[tuple[str, int]],
    target_amount: int,
    target_with_change: int,
    accumulated_amount: int,
) -> tuple[set[str], int, bool]:
    """Pick UTxOs to meet or exceed the target amount while prioritizing defragmentation.

    Args:
        utxos (list of tuples): A list of tuples (utxo_id, coin_amount).
        target_amount (int): The desired amount.
        target_with_change (int): The desired amount with minimal change.
        accumulated_amount (int): The accumulated amount.

    Returns:
        tuple: A tuple with selected UTxO IDs, accumulated amount and a bool indicating if the
            desired amount was met.
    """
    # Sort UTxOs by amount in ascending order
    sorted_utxos = sorted(enumerate(utxos), key=lambda x: x[1][1])  # Keep original indices
    selected_indices = set()
    picked_utxos = set()

    # Step 1: Select up to 10 smallest UTxOs
    for i, (utxo_id, coin_amount) in sorted_utxos[:10]:
        picked_utxos.add(utxo_id)
        selected_indices.add(i)
        accumulated_amount += coin_amount

        # If we were able to collect exact amount, no change is needed
        if accumulated_amount == target_amount:
            return picked_utxos, accumulated_amount, True
        # Make sure the change is higher than `_min_change_value`
        if accumulated_amount >= target_with_change:
            return picked_utxos, accumulated_amount, True

    # Step 2: If target is not met, select UTxO closest to remaining amount
    while accumulated_amount < target_with_change:
        # If we were able to collect exact amount, no change is needed
        if accumulated_amount == target_amount:
            return picked_utxos, accumulated_amount, True

        # We target exact amount, but if we are already over it, we need at least additional
        # `_min_change_value` for change.
        if accumulated_amount > target_amount:
            remaining_amount = target_with_change - accumulated_amount
        else:
            remaining_amount = target_amount - accumulated_amount

        # Find the index of the UTxO closest to the remaining amount
        closest_index = min(
            (i for i, _ in sorted_utxos if i not in selected_indices),
            key=lambda i: abs(utxos[i][1] - remaining_amount),
            default=None,
        )

        # If all UTxOs have been considered, the target was not met
        if closest_index is None:
            return picked_utxos, accumulated_amount, False

        # Select the closest UTxO
        utxo_id, coin_amount = utxos[closest_index]
        picked_utxos.add(utxo_id)
        selected_indices.add(closest_index)
        accumulated_amount += coin_amount

    return picked_utxos, accumulated_amount, True


def _select_utxos_per_coin(
    coin_txins: dict[str, int],
    coin: str,
    target_amount: int,
    target_with_change: int,
    already_selected_utxos: set[str],
) -> set[str]:
    """Select UTxOs for a given coin so their total combined amount >= `amount`."""
    selected_utxos, accumulated_amount, target_met = _pick_coins_from_already_selected_utxos(
        coin_txins=coin_txins,
        already_selected_utxos=already_selected_utxos,
        target_amount=target_amount,
        target_with_change=target_with_change,
    )

    # Pick more UTxOs if the amount is not satisfied yet
    if not target_met:
        ids_and_amounts = [(i, a) for i, a in coin_txins.items() if i not in already_selected_utxos]
        more_utxos, _, target_met = _pick_utxos_with_defragmentation(
            utxos=ids_and_amounts,
            target_amount=target_amount,
            target_with_change=target_with_change,
            accumulated_amount=accumulated_amount,
        )
        selected_utxos.update(more_utxos)

    if not target_met:
        LOGGER.warning(
            f"Could not meet target amount {target_amount} for coin '{coin}' with the given UTxOs."
        )

    return selected_utxos


def _select_utxos(
    txins_by_coin_and_id: dict[str, dict[str, int]],
    txouts_passed_db: dict[str, list[structs.TxOut]],
    txouts_mint_db: dict[str, list[structs.TxOut]],
    fee: int,
    withdrawals: structs.OptionalTxOuts,
    min_change_value: int,
    deposit: int = 0,
    treasury_donation: int = 0,
) -> set[str]:
    """Select UTxOs that can satisfy all outputs, deposits and fee.

    Return IDs of selected UTxOs.
    """
    utxo_ids: set[str] = set()

    # Iterate over coins both in txins and txouts
    for coin in set(txins_by_coin_and_id).union(txouts_passed_db).union(txouts_mint_db):
        coin_txins = txins_by_coin_and_id.get(coin) or {}
        coin_txouts = txouts_passed_db.get(coin) or []

        total_output_amount = functools.reduce(lambda x, y: x + y.amount, coin_txouts, 0)

        if coin == consts.DEFAULT_COIN:
            # The value "-1" means all available funds
            max_index = [idx for idx, val in enumerate(coin_txouts) if val.amount == -1]
            if max_index:
                utxo_ids.update(r for r in coin_txins)
                continue

            tx_fee = max(1, fee)
            funds_needed = total_output_amount + tx_fee + deposit + treasury_donation
            total_withdrawals_amount = functools.reduce(lambda x, y: x + y.amount, withdrawals, 0)
            # Fee needs an input, even if withdrawal would cover all needed funds
            input_funds_needed = max(funds_needed - total_withdrawals_amount, tx_fee)
            # `_min_change_value` applies only to ADA
            target_with_change = input_funds_needed + min_change_value
        else:
            coin_txouts_minted = txouts_mint_db.get(coin) or []
            total_minted_amount = functools.reduce(lambda x, y: x + y.amount, coin_txouts_minted, 0)
            # In case of token burning, `total_minted_amount` might be negative.
            # Try to collect enough funds to satisfy both token burning and token
            # transfers, even though there might be an overlap.
            input_funds_needed = total_output_amount - total_minted_amount
            target_with_change = input_funds_needed

        if input_funds_needed:
            utxo_ids.update(
                _select_utxos_per_coin(
                    coin_txins=txins_by_coin_and_id.get(coin) or {},
                    coin=coin,
                    target_amount=input_funds_needed,
                    target_with_change=target_with_change,
                    already_selected_utxos=utxo_ids,
                )
            )

    return utxo_ids


def _balance_txouts(
    change_address: str,
    txouts: structs.OptionalTxOuts,
    txins_db: dict[str, list[structs.UTXOData]],
    txouts_passed_db: dict[str, list[structs.TxOut]],
    txouts_mint_db: dict[str, list[structs.TxOut]],
    fee: int,
    withdrawals: structs.OptionalTxOuts,
    deposit: int = 0,
    treasury_donation: int = 0,
    skip_asset_balancing: bool = False,
) -> list[structs.TxOut]:
    """Balance the transaction by adding change output for each coin."""
    # Records for burning tokens, i.e. records with negative amount, are not allowed in `txouts`
    burning_txouts = [r for r in txouts if r.amount < 0 and r.coin != consts.DEFAULT_COIN]
    if burning_txouts:
        msg = f"Token burning is not allowed in txouts: {burning_txouts}"
        raise AssertionError(msg)

    # Filter out negative amounts (-1 "max" amounts)
    txouts_result = [r for r in txouts if r.amount > 0]

    if skip_asset_balancing:
        # Balancing is done elsewhere (by the `transaction build` command)
        return txouts_result

    # Iterate over coins both in txins and txouts
    for coin in set(txins_db).union(txouts_passed_db).union(txouts_mint_db):
        max_address = None
        change = 0

        coin_txins = txins_db.get(coin) or []
        coin_txouts = txouts_passed_db.get(coin) or []

        total_input_amount = functools.reduce(lambda x, y: x + y.amount, coin_txins, 0)

        if coin == consts.DEFAULT_COIN:
            # The value "-1" means all available funds
            max_index = [idx for idx, val in enumerate(coin_txouts) if val.amount == -1]
            if len(max_index) > 1:
                msg = "Cannot send all remaining funds to more than one address."
                raise AssertionError(msg)
            if max_index:
                # Remove the "-1" record and get its address
                max_address = coin_txouts.pop(max_index[0]).address

            total_output_amount = functools.reduce(lambda x, y: x + y.amount, coin_txouts, 0)
            tx_fee = max(0, fee)
            total_withdrawals_amount = functools.reduce(lambda x, y: x + y.amount, withdrawals, 0)
            funds_available = total_input_amount + total_withdrawals_amount
            funds_needed = total_output_amount + tx_fee + deposit + treasury_donation
            change = funds_available - funds_needed
            if change < 0:
                LOGGER.error(
                    "Not enough funds to make the transaction - "
                    f"available: {funds_available}; needed: {funds_needed}"
                )
        else:
            coin_txouts_minted = txouts_mint_db.get(coin) or []
            total_output_amount = functools.reduce(lambda x, y: x + y.amount, coin_txouts, 0)
            total_minted_amount = functools.reduce(lambda x, y: x + y.amount, coin_txouts_minted, 0)
            funds_available = total_input_amount + total_minted_amount
            change = funds_available - total_output_amount
            if change < 0:
                LOGGER.error(
                    f"Amount of coin `{coin}` is not sufficient - "
                    f"available: {funds_available}; needed: {total_output_amount}"
                )

        if change > 0:
            txouts_result.append(
                structs.TxOut(address=(max_address or change_address), amount=change, coin=coin)
            )

    return txouts_result


def _resolve_withdrawals(
    clusterlib_obj: "itp.ClusterLib", withdrawals: list[structs.TxOut]
) -> list[structs.TxOut]:
    """Return list of resolved reward withdrawals.

    The `structs.TxOut.amount` can be '-1', meaning all available funds.

    Args:
        clusterlib_obj: An instance of `ClusterLib`.
        withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals.

    Returns:
        List[structs.TxOut]: A list of `TxOuts`, specifying resolved reward withdrawals.
    """
    resolved_withdrawals = []
    for rec in withdrawals:
        # The amount with value "-1" means all available balance
        if rec.amount == -1:
            balance = clusterlib_obj.g_query.get_stake_addr_info(rec.address).reward_account_balance
            resolved_withdrawals.append(structs.TxOut(address=rec.address, amount=balance))
        else:
            resolved_withdrawals.append(rec)

    return resolved_withdrawals


def _get_withdrawals(
    clusterlib_obj: "itp.ClusterLib",
    withdrawals: structs.OptionalTxOuts,
    script_withdrawals: structs.OptionalScriptWithdrawals,
) -> tuple[structs.OptionalTxOuts, structs.OptionalScriptWithdrawals, structs.OptionalTxOuts]:
    """Return tuple of resolved withdrawals.

    Return simple withdrawals, script withdrawals, combination of all withdrawals Tx outputs.
    """
    withdrawals = withdrawals and _resolve_withdrawals(
        clusterlib_obj=clusterlib_obj, withdrawals=withdrawals
    )
    script_withdrawals = [
        dataclasses.replace(
            s, txout=_resolve_withdrawals(clusterlib_obj=clusterlib_obj, withdrawals=[s.txout])[0]
        )
        for s in script_withdrawals
    ]
    withdrawals_txouts = [*withdrawals, *[s.txout for s in script_withdrawals]]
    return withdrawals, script_withdrawals, withdrawals_txouts


def _get_reference_txins(
    readonly_reference_txins: structs.OptionalUTXOData,
    script_txins: structs.OptionalScriptTxIn,
    mint: structs.OptionalMint,
    complex_certs: structs.OptionalScriptCerts,
    script_withdrawals: structs.OptionalScriptWithdrawals,
) -> list[structs.UTXOData]:
    """Get list of reference txins."""
    script_ref_txins = [
        r.reference_txin
        for r in (
            *script_txins,
            *mint,
            *complex_certs,
            *script_withdrawals,
        )
        if r.reference_txin
    ]

    return [*readonly_reference_txins, *script_ref_txins]


def _get_txin_strings(
    txins: structs.OptionalUTXOData, script_txins: structs.OptionalScriptTxIn
) -> set[str]:
    """Get list of txin strings for normal (non-script) inputs."""
    # Filter out duplicate txins
    txins_utxos = {f"{x.utxo_hash}#{x.utxo_ix}" for x in txins}

    # Assume that all plutus txin records are for the same UTxO and use the first one
    plutus_txins_utxos = {
        f"{x.txins[0].utxo_hash}#{x.txins[0].utxo_ix}" for x in script_txins if x.txins
    }

    # Remove plutus txin records from normal txins
    txins_combined = txins_utxos.difference(plutus_txins_utxos)

    return txins_combined


def _get_txout_plutus_args(txout: structs.TxOut) -> list[str]:  # noqa: C901
    txout_args = []

    # Add datum arguments
    if txout.datum_hash:
        txout_args = [
            "--tx-out-datum-hash",
            str(txout.datum_hash),
        ]
    elif txout.datum_hash_file:
        txout_args = [
            "--tx-out-datum-hash-file",
            str(txout.datum_hash_file),
        ]
    elif txout.datum_hash_cbor_file:
        txout_args = [
            "--tx-out-datum-hash-cbor-file",
            str(txout.datum_hash_cbor_file),
        ]
    elif txout.datum_hash_value:
        txout_args = [
            "--tx-out-datum-hash-value",
            str(txout.datum_hash_value),
        ]
    elif txout.datum_embed_file:
        txout_args = [
            "--tx-out-datum-embed-file",
            str(txout.datum_embed_file),
        ]
    elif txout.datum_embed_cbor_file:
        txout_args = [
            "--tx-out-datum-embed-cbor-file",
            str(txout.datum_embed_cbor_file),
        ]
    elif txout.datum_embed_value:
        txout_args = [
            "--tx-out-datum-embed-value",
            str(txout.datum_embed_value),
        ]
    elif txout.inline_datum_file:
        txout_args = [
            "--tx-out-inline-datum-file",
            str(txout.inline_datum_file),
        ]
    elif txout.inline_datum_cbor_file:
        txout_args = [
            "--tx-out-inline-datum-cbor-file",
            str(txout.inline_datum_cbor_file),
        ]
    elif txout.inline_datum_value:
        txout_args = [
            "--tx-out-inline-datum-value",
            str(txout.inline_datum_value),
        ]

    # Add reference script arguments
    if txout.reference_script_file:
        txout_args.extend(
            [
                "--tx-out-reference-script-file",
                str(txout.reference_script_file),
            ]
        )

    return txout_args


def get_joined_txouts(
    txouts: list[structs.TxOut],
) -> list[list[structs.TxOut]]:
    """Return list of joined TxOuts."""
    txouts_by_eutxo_attrs: dict[str, list[structs.TxOut]] = {}
    joined_txouts: list[list[structs.TxOut]] = []

    # Aggregate TX outputs by address, datum and reference script
    for rec in txouts:
        datum_src = str(
            rec.datum_hash
            or rec.datum_hash_file
            or rec.datum_hash_cbor_file
            or rec.datum_hash_value
            or rec.datum_embed_file
            or rec.datum_embed_cbor_file
            or rec.datum_embed_value
        )

        inline_datum_src = str(
            rec.inline_datum_file or rec.inline_datum_cbor_file or rec.inline_datum_value
        )

        eutxo_attrs = f"{rec.address}::{datum_src}::{inline_datum_src}::{rec.reference_script_file}"

        if eutxo_attrs not in txouts_by_eutxo_attrs:
            txouts_by_eutxo_attrs[eutxo_attrs] = []
        txouts_by_eutxo_attrs[eutxo_attrs].append(rec)

    # Join txouts with the same address, datum and reference script
    for txouts_list in txouts_by_eutxo_attrs.values():
        # Create single `TxOut` record with sum of amounts per coin
        txouts_by_coin: dict[str, tuple[structs.TxOut, list[int]]] = {}
        for ar in txouts_list:
            if ar.coin in txouts_by_coin:
                txouts_by_coin[ar.coin][1].append(ar.amount)
            else:
                txouts_by_coin[ar.coin] = (ar, [ar.amount])
        # The tuple for each coin is `("one of the original TxOuts", "list of amounts")`.
        # All the `TxOut` values except of amount are the same in this loop, so we can
        # take the original `TxOut` and replace `amount` with sum of all amounts.
        sum_txouts = [dataclasses.replace(r[0], amount=sum(r[1])) for r in txouts_by_coin.values()]

        joined_txouts.append(sum_txouts)

    return joined_txouts


def _join_txouts(
    txouts: list[structs.TxOut],
) -> tuple[list[str], list[structs.TxOut], int]:
    txout_args: list[str] = []
    joined_txouts = get_joined_txouts(txouts=txouts)
    for joined_recs in joined_txouts:
        amounts = [
            f"{r.amount} {r.coin if r.coin != consts.DEFAULT_COIN else ''}".rstrip()
            for r in joined_recs
        ]
        amounts_joined = "+".join(amounts)

        txout_args.extend(["--tx-out", f"{joined_recs[0].address}+{amounts_joined}"])
        txout_args.extend(_get_txout_plutus_args(txout=joined_recs[0]))

    joined_txouts_flat = list(itertools.chain.from_iterable(joined_txouts))
    return txout_args, joined_txouts_flat, len(joined_txouts)


def _list_txouts(txouts: list[structs.TxOut]) -> list[str]:
    txout_args: list[str] = []

    for rec in txouts:
        txout_args.extend(
            [
                "--tx-out",
                f"{rec.address}+{rec.amount} "
                f"{rec.coin if rec.coin != consts.DEFAULT_COIN else ''}".rstrip(),
            ]
        )
        txout_args.extend(_get_txout_plutus_args(txout=rec))

    return txout_args


def _get_return_collateral_txout_args(txouts: structs.OptionalTxOuts) -> list[str]:
    if not txouts:
        return []

    addresses = {t.address for t in txouts}
    if len(addresses) > 1:
        msg = "Accepts `txouts` only for single address."
        raise AssertionError(msg)

    txout_records = [
        f"{t.amount} {t.coin if t.coin != consts.DEFAULT_COIN else ''}".rstrip() for t in txouts
    ]
    address_value = "{}+{}".format(txouts[0].address, "+".join(txout_records))
    txout_args = ["--tx-out-return-collateral", address_value]

    return txout_args


def _process_txouts(
    txouts: list[structs.TxOut], join_txouts: bool
) -> tuple[list[str], list[structs.TxOut], int]:
    if join_txouts:
        return _join_txouts(txouts=txouts)
    return _list_txouts(txouts=txouts), txouts, len(txouts)


def _get_tx_ins_outs(
    clusterlib_obj: "itp.ClusterLib",
    src_address: str,
    tx_files: structs.TxFiles,
    txins: structs.OptionalUTXOData = (),
    txouts: structs.OptionalTxOuts = (),
    fee: int = 0,
    deposit: int | None = None,
    treasury_donation: int | None = None,
    withdrawals: structs.OptionalTxOuts = (),
    mint_txouts: structs.OptionalTxOuts = (),
    src_addr_utxos: list[structs.UTXOData] | None = None,
    skip_asset_balancing: bool = False,
) -> tuple[list[structs.UTXOData], list[structs.TxOut]]:
    """Return list of transaction's inputs and outputs.

    Args:
        clusterlib_obj: An instance of `ClusterLib`.
        src_address: An address used for fee and inputs (if inputs not specified by `txins`).
        tx_files: A `structs.TxFiles` data container containing files needed for the transaction.
        txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
        txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
        fee: A fee amount (optional).
        deposit: A deposit amount needed by the transaction (optional).
        treasury_donation: A donation to the treasury to perform (optional).
        withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
        mint_txouts: A list (iterable) of `TxOuts`, specifying minted tokens (optional).
        src_addr_utxos: A list of UTxOs for the source address (optional).
        skip_asset_balancing: A bool indicating if assets balancing should be skipped
            (`build` command balance the assets automatically in newer versions).

    Returns:
        Tuple[list, list]: A tuple of list of transaction inputs and list of transaction
            outputs.
    """
    txouts_passed_db: dict[str, list[structs.TxOut]] = _organize_tx_ins_outs_by_coin(txouts)
    txouts_mint_db: dict[str, list[structs.TxOut]] = _organize_tx_ins_outs_by_coin(mint_txouts)
    outcoins_all = {consts.DEFAULT_COIN, *txouts_mint_db.keys(), *txouts_passed_db.keys()}
    outcoins_passed = [consts.DEFAULT_COIN, *txouts_passed_db.keys()]

    txins_all = list(txins)
    if not txins_all:
        # No txins were provided, so we'll select them from the source address
        address_utxos = src_addr_utxos or clusterlib_obj.g_query.get_utxo(address=src_address)
        if not address_utxos:
            msg = f"No UTxO returned for '{src_address}'."
            raise exceptions.CLIError(msg)
        txins_all = _get_usable_utxos(address_utxos=address_utxos, coins=outcoins_all)

    if not txins_all:
        msg = "No input UTxO."
        raise exceptions.CLIError(msg)

    txins_by_coin_and_id = _organize_utxos_by_coin_and_id(txins_all)

    # All output coins, except those minted by this transaction, need to be present in
    # transaction inputs
    if not set(outcoins_passed).difference(txouts_mint_db).issubset(txins_by_coin_and_id):
        msg = "Not all output coins are present in input UTxOs."
        raise exceptions.CLIError(msg)

    tx_deposit = (
        clusterlib_obj.g_transaction.get_tx_deposit(tx_files=tx_files)
        if deposit is None
        else deposit
    )

    tx_treasury_donation = treasury_donation if treasury_donation is not None else 0

    if txins:
        # Don't touch txins that were passed to the function
        txins_filtered = txins_all
        txins_db_filtered = _organize_tx_ins_outs_by_coin(txins_all)
    else:
        # Select only UTxOs that are needed to satisfy all outputs, deposits and fee
        selected_utxo_ids = _select_utxos(
            txins_by_coin_and_id=txins_by_coin_and_id,
            txouts_passed_db=txouts_passed_db,
            txouts_mint_db=txouts_mint_db,
            fee=fee,
            withdrawals=withdrawals,
            min_change_value=clusterlib_obj._min_change_value,
            deposit=tx_deposit,
            treasury_donation=tx_treasury_donation,
        )
        txins_by_id: dict[str, list[structs.UTXOData]] = _organize_utxos_by_id(txins_all)
        _txins_filtered = [utxo for uid, utxo in txins_by_id.items() if uid in selected_utxo_ids]

        txins_filtered = list(itertools.chain.from_iterable(_txins_filtered))
        txins_db_filtered = _organize_tx_ins_outs_by_coin(txins_filtered)

    if not txins_filtered:
        msg = "Cannot build transaction, empty `txins`."
        raise exceptions.CLIError(msg)

    # Balance the transaction
    txouts_balanced = _balance_txouts(
        # Return change to `src_address`.
        # When using `build_tx`, Lovelace change is returned to `change_address` (this is handled
        # automatically by `transaction build`) and only tokens change is returned to
        # `src_address`. It is up to user to specify Lovelace output for `src_address` with high
        # enough Lovelace value when token change is needed and `change_address` differs from
        # `src_address`.
        change_address=src_address,
        txouts=txouts,
        txins_db=txins_db_filtered,
        txouts_passed_db=txouts_passed_db,
        txouts_mint_db=txouts_mint_db,
        fee=fee,
        withdrawals=withdrawals,
        deposit=tx_deposit,
        treasury_donation=tx_treasury_donation,
        skip_asset_balancing=skip_asset_balancing,
    )

    return txins_filtered, txouts_balanced


def collect_data_for_build(
    clusterlib_obj: "itp.ClusterLib",
    src_address: str,
    txins: structs.OptionalUTXOData = (),
    txouts: structs.OptionalTxOuts = (),
    script_txins: structs.OptionalScriptTxIn = (),
    mint: structs.OptionalMint = (),
    tx_files: structs.TxFiles | None = None,
    complex_certs: structs.OptionalScriptCerts = (),
    complex_proposals: structs.OptionalScriptProposals = (),
    fee: int = 0,
    withdrawals: structs.OptionalTxOuts = (),
    script_withdrawals: structs.OptionalScriptWithdrawals = (),
    deposit: int | None = None,
    treasury_donation: int | None = None,
    src_addr_utxos: list[structs.UTXOData] | None = None,
    skip_asset_balancing: bool = False,
) -> structs.DataForBuild:
    """Collect data (txins, txouts, withdrawals) needed for building a transaction.

    Args:
        clusterlib_obj: An instance of `ClusterLib`.
        src_address: An address used for fee and inputs (if inputs not specified by `txins`).
        txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
        txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
        script_txins: An iterable of `ScriptTxIn`, specifying input script UTxOs (optional).
        mint: An iterable of `Mint`, specifying script minting data (optional).
        tx_files: A `structs.TxFiles` data container containing files needed for the transaction
            (optional).
        complex_certs: An iterable of `ComplexCert`, specifying certificates script data
            (optional).
        complex_proposals: An iterable of `ComplexProposal`, specifying proposals script data
            (optional).
        fee: A fee amount (optional).
        withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
        script_withdrawals: An iterable of `ScriptWithdrawal`, specifying withdrawal script
            data (optional).
        deposit: A deposit amount needed by the transaction (optional).
        treasury_donation: A donation to the treasury to perform (optional).
        src_addr_utxos: A list of UTxOs for the source address (optional).
        skip_asset_balancing: A bool indicating if assets balancing should be skipped
            (`build` command balance the assets automatically in newer versions).

    Returns:
        structs.DataForBuild: A data container with data for build(-raw) commands.
    """
    tx_files = tx_files or structs.TxFiles()

    withdrawals, script_withdrawals, withdrawals_txouts = _get_withdrawals(
        clusterlib_obj=clusterlib_obj,
        withdrawals=withdrawals,
        script_withdrawals=script_withdrawals,
    )

    script_txins_records = list(itertools.chain.from_iterable(r.txins for r in script_txins))

    script_addresses = {r.address for r in script_txins_records}
    if src_address in script_addresses:
        msg = "Source address cannot be a script address."
        raise AssertionError(msg)

    # Combine txins and make sure we have enough funds to satisfy all txouts
    combined_txins = [
        *txins,
        *script_txins_records,
    ]
    mint_txouts = list(itertools.chain.from_iterable(m.txouts for m in mint))
    combined_tx_files = dataclasses.replace(
        tx_files,
        certificate_files=[
            *tx_files.certificate_files,
            *[c.certificate_file for c in complex_certs],
        ],
        proposal_files=[
            *tx_files.proposal_files,
            *[p.proposal_file for p in complex_proposals],
        ],
    )
    txins_copy, txouts_copy = _get_tx_ins_outs(
        clusterlib_obj=clusterlib_obj,
        src_address=src_address,
        tx_files=combined_tx_files,
        txins=combined_txins,
        txouts=txouts,
        fee=fee,
        deposit=deposit,
        treasury_donation=treasury_donation,
        withdrawals=withdrawals_txouts,
        mint_txouts=mint_txouts,
        src_addr_utxos=src_addr_utxos,
        skip_asset_balancing=skip_asset_balancing,
    )

    payment_txins = txins or txins_copy
    # Don't include script txins in list of payment txins
    if script_txins_records:
        payment_txins = txins or []

    return structs.DataForBuild(
        txins=payment_txins,
        txouts=txouts_copy,
        withdrawals=withdrawals,
        script_withdrawals=script_withdrawals,
    )


def get_utxo(
    utxo_dict: dict,
    address: str = "",
    coins: itp.UnpackableSequence = (),
) -> list[structs.UTXOData]:
    """Return UTxO info for payment address.

    Args:
        utxo_dict: A JSON output of `query utxo`.
        address: A payment address.
        coins: A list (iterable) of coin names (asset IDs).

    Returns:
        List[structs.UTXOData]: A list of UTxO data.
    """
    utxo = []
    for utxo_rec, utxo_data in utxo_dict.items():
        utxo_hash, utxo_ix = utxo_rec.split("#")
        utxo_address = utxo_data.get("address") or ""
        addr_data = utxo_data["value"]
        datum_hash = utxo_data.get("data") or utxo_data.get("datumhash") or ""
        inline_datum_hash = utxo_data.get("inlineDatumhash") or ""
        inline_datum = utxo_data.get("inlineDatum")
        reference_script = utxo_data.get("referenceScript")

        for policyid, coin_data in addr_data.items():
            if policyid == consts.DEFAULT_COIN:
                utxo.append(
                    structs.UTXOData(
                        utxo_hash=utxo_hash,
                        utxo_ix=int(utxo_ix),
                        amount=coin_data,
                        address=address or utxo_address,
                        coin=consts.DEFAULT_COIN,
                        datum_hash=datum_hash,
                        inline_datum_hash=inline_datum_hash,
                        inline_datum=inline_datum,
                        reference_script=reference_script,
                    )
                )
                continue

            # Coin data used to be a dict, now it is a list
            try:
                coin_iter = coin_data.items()
            except AttributeError:
                coin_iter = coin_data

            for asset_name, amount in coin_iter:
                decoded_coin = ""
                if asset_name:
                    with contextlib.suppress(Exception):
                        decoded_name = base64.b16decode(asset_name.encode(), casefold=True).decode(
                            "utf-8"
                        )
                        decoded_coin = f"{policyid}.{decoded_name}"
                else:
                    decoded_coin = policyid

                utxo.append(
                    structs.UTXOData(
                        utxo_hash=utxo_hash,
                        utxo_ix=int(utxo_ix),
                        amount=amount,
                        address=address or utxo_address,
                        coin=f"{policyid}.{asset_name}" if asset_name else policyid,
                        decoded_coin=decoded_coin,
                        datum_hash=datum_hash,
                        inline_datum_hash=inline_datum_hash,
                        inline_datum=inline_datum,
                        reference_script=reference_script,
                    )
                )

    if coins:
        filtered_utxo = [u for u in utxo if u.coin in coins]
        return filtered_utxo

    return utxo


def calculate_utxos_balance(
    utxos: list[structs.UTXOData] | list[structs.TxOut],
    coin: str = consts.DEFAULT_COIN,
) -> int:
    """Calculate sum of UTxO balances.

    Args:
        utxos: A list of UTxO data (either `structs.UTXOData` or `structs.TxOut`).
        coin: A coin name (asset IDs).

    Returns:
        int: A total balance.
    """
    filtered_utxos = [u for u in utxos if u.coin == coin]
    address_balance = functools.reduce(lambda x, y: x + y.amount, filtered_utxos, 0)
    return int(address_balance)


def filter_utxo_with_highest_amount(
    utxos: list[structs.UTXOData],
    coin: str = consts.DEFAULT_COIN,
) -> structs.UTXOData:
    """Return data for UTxO with the highest amount.

    Args:
        utxos: A list of UTxO data.
        coin: A coin name (asset IDs).

    Returns:
        structs.UTXOData: A UTxO record with the highest amount.
    """
    filtered_utxos = [u for u in utxos if u.coin == coin]
    highest_amount_rec = max(filtered_utxos, key=lambda x: x.amount)
    return highest_amount_rec


def filter_utxos(
    utxos: list[structs.UTXOData],
    utxo_hash: str = "",
    utxo_ix: int | None = None,
    amount: int | None = None,
    address: str = "",
    coin: str = "",
    datum_hash: str = "",
    inline_datum_hash: str = "",
) -> list[structs.UTXOData]:
    """Get UTxO records that match given filtering criteria.

    Args:
        utxos: A list of UTxO data.
        utxo_hash: A transaction identifier (optional).
        utxo_ix: A UTxO index (optional).
        amount: An amount of coin (optional).
        address: A payment address (optional).
        coin: A coin name (asset ID; optional).
        datum_hash: A datum hash (optional).
        inline_datum_hash: An inline datum hash (optional).

    Returns:
        structs.UTXOData: UTxO records that match given filtering criteria.
    """
    filtered_utxos = []

    for u in utxos:
        if utxo_hash and u.utxo_hash != utxo_hash:
            continue
        if utxo_ix is not None and utxo_ix != u.utxo_ix:
            continue
        if amount is not None and amount != u.amount:
            continue
        if address and u.address != address:
            continue
        if coin and u.coin != coin:
            continue
        if datum_hash and u.datum_hash != datum_hash:
            continue
        if inline_datum_hash and u.inline_datum_hash != inline_datum_hash:
            continue
        filtered_utxos.append(u)

    return filtered_utxos


def _get_script_args(  # noqa: C901
    script_txins: structs.OptionalScriptTxIn,
    mint: structs.OptionalMint,
    complex_certs: structs.OptionalScriptCerts,
    complex_proposals: structs.OptionalScriptProposals,
    script_withdrawals: structs.OptionalScriptWithdrawals,
    script_votes: structs.OptionalScriptVotes,
    with_execution_units: bool = True,
) -> list[str]:
    grouped_args: list[str] = []
    collaterals_all = set()

    # Spending
    for tin in script_txins:
        if tin.txins:
            grouped_args.extend(
                [
                    "--tx-in",
                    # Assume that all txin records are for the same UTxO and use the first one
                    f"{tin.txins[0].utxo_hash}#{tin.txins[0].utxo_ix}",
                ]
            )
        tin_collaterals = {f"{c.utxo_hash}#{c.utxo_ix}" for c in tin.collaterals}
        collaterals_all.update(tin_collaterals)

        if tin.script_file:
            grouped_args.extend(
                [
                    "--tx-in-script-file",
                    str(tin.script_file),
                ]
            )

            if with_execution_units and tin.execution_units:
                grouped_args.extend(
                    [
                        "--tx-in-execution-units",
                        f"({tin.execution_units[0]},{tin.execution_units[1]})",
                    ]
                )

            if tin.datum_file:
                grouped_args.extend(["--tx-in-datum-file", str(tin.datum_file)])
            if tin.datum_cbor_file:
                grouped_args.extend(["--tx-in-datum-cbor-file", str(tin.datum_cbor_file)])
            if tin.datum_value:
                grouped_args.extend(["--tx-in-datum-value", str(tin.datum_value)])
            if tin.inline_datum_present:
                grouped_args.append("--tx-in-inline-datum-present")
            if tin.redeemer_file:
                grouped_args.extend(["--tx-in-redeemer-file", str(tin.redeemer_file)])
            if tin.redeemer_cbor_file:
                grouped_args.extend(["--tx-in-redeemer-cbor-file", str(tin.redeemer_cbor_file)])
            if tin.redeemer_value:
                grouped_args.extend(["--tx-in-redeemer-value", str(tin.redeemer_value)])

        if tin.reference_txin:
            tin_reference_txin_id = f"{tin.reference_txin.utxo_hash}#{tin.reference_txin.utxo_ix}"
            tin_reference_type = tin.reference_type or consts.ScriptTypes.PLUTUS_V2

            if tin_reference_type in (consts.ScriptTypes.SIMPLE_V1, consts.ScriptTypes.SIMPLE_V2):
                grouped_args.extend(
                    [
                        "--simple-script-tx-in-reference",
                        tin_reference_txin_id,
                    ]
                )
            else:
                grouped_args.extend(
                    [
                        "--spending-tx-in-reference",
                        tin_reference_txin_id,
                    ]
                )

            if tin.reference_type == consts.ScriptTypes.PLUTUS_V2:
                grouped_args.append("--spending-plutus-script-v2")
            elif tin.reference_type == consts.ScriptTypes.PLUTUS_V3:
                grouped_args.append("--spending-plutus-script-v3")

            if with_execution_units and tin.execution_units:
                grouped_args.extend(
                    [
                        "--spending-reference-tx-in-execution-units",
                        f"({tin.execution_units[0]},{tin.execution_units[1]})",
                    ]
                )

            if tin.datum_file:
                grouped_args.extend(["--spending-reference-tx-in-datum-file", str(tin.datum_file)])
            if tin.datum_cbor_file:
                grouped_args.extend(
                    ["--spending-reference-tx-in-datum-cbor-file", str(tin.datum_cbor_file)]
                )
            if tin.datum_value:
                grouped_args.extend(
                    ["--spending-reference-tx-in-datum-value", str(tin.datum_value)]
                )
            if tin.inline_datum_present:
                grouped_args.append("--spending-reference-tx-in-inline-datum-present")
            if tin.redeemer_file:
                grouped_args.extend(
                    ["--spending-reference-tx-in-redeemer-file", str(tin.redeemer_file)]
                )
            if tin.redeemer_cbor_file:
                grouped_args.extend(
                    ["--spending-reference-tx-in-redeemer-cbor-file", str(tin.redeemer_cbor_file)]
                )
            if tin.redeemer_value:
                grouped_args.extend(
                    ["--spending-reference-tx-in-redeemer-value", str(tin.redeemer_value)]
                )

    # Minting
    for mrec in mint:
        mrec_collaterals = {f"{c.utxo_hash}#{c.utxo_ix}" for c in mrec.collaterals}
        collaterals_all.update(mrec_collaterals)

        if mrec.script_file:
            grouped_args.extend(
                [
                    "--mint-script-file",
                    str(mrec.script_file),
                ]
            )

            if with_execution_units and mrec.execution_units:
                grouped_args.extend(
                    [
                        "--mint-execution-units",
                        f"({mrec.execution_units[0]},{mrec.execution_units[1]})",
                    ]
                )

            if mrec.redeemer_file:
                grouped_args.extend(["--mint-redeemer-file", str(mrec.redeemer_file)])
            if mrec.redeemer_cbor_file:
                grouped_args.extend(["--mint-redeemer-cbor-file", str(mrec.redeemer_cbor_file)])
            if mrec.redeemer_value:
                grouped_args.extend(["--mint-redeemer-value", str(mrec.redeemer_value)])

        if mrec.reference_txin:
            mrec_reference_txin_id = (
                f"{mrec.reference_txin.utxo_hash}#{mrec.reference_txin.utxo_ix}"
            )
            mrec_reference_type = mrec.reference_type or consts.ScriptTypes.PLUTUS_V2

            if mrec_reference_type in (consts.ScriptTypes.SIMPLE_V1, consts.ScriptTypes.SIMPLE_V2):
                grouped_args.extend(
                    [
                        "--simple-minting-script-tx-in-reference",
                        mrec_reference_txin_id,
                    ]
                )
            else:
                grouped_args.extend(
                    [
                        "--mint-tx-in-reference",
                        mrec_reference_txin_id,
                    ]
                )
            if mrec_reference_type == consts.ScriptTypes.PLUTUS_V2:
                grouped_args.append("--mint-plutus-script-v2")
            elif mrec_reference_type == consts.ScriptTypes.PLUTUS_V3:
                grouped_args.append("--mint-plutus-script-v3")

            if with_execution_units and mrec.execution_units:
                grouped_args.extend(
                    [
                        "--mint-reference-tx-in-execution-units",
                        f"({mrec.execution_units[0]},{mrec.execution_units[1]})",
                    ]
                )

            if mrec.redeemer_file:
                grouped_args.extend(
                    ["--mint-reference-tx-in-redeemer-file", str(mrec.redeemer_file)]
                )
            if mrec.redeemer_cbor_file:
                grouped_args.extend(
                    ["--mint-reference-tx-in-redeemer-cbor-file", str(mrec.redeemer_cbor_file)]
                )
            if mrec.redeemer_value:
                grouped_args.extend(
                    ["--mint-reference-tx-in-redeemer-value", str(mrec.redeemer_value)]
                )
            if mrec.policyid:
                grouped_args.extend(["--policy-id", str(mrec.policyid)])

    # Certificates
    for crec in complex_certs:
        crec_collaterals = {f"{c.utxo_hash}#{c.utxo_ix}" for c in crec.collaterals}
        collaterals_all.update(crec_collaterals)
        grouped_args.extend(
            [
                "--certificate-file",
                str(crec.certificate_file),
            ]
        )

        if crec.script_file:
            grouped_args.extend(["--certificate-script-file", str(crec.script_file)])

            if with_execution_units and crec.execution_units:
                grouped_args.extend(
                    [
                        "--certificate-execution-units",
                        f"({crec.execution_units[0]},{crec.execution_units[1]})",
                    ]
                )

            if crec.redeemer_file:
                grouped_args.extend(["--certificate-redeemer-file", str(crec.redeemer_file)])
            if crec.redeemer_cbor_file:
                grouped_args.extend(
                    ["--certificate-redeemer-cbor-file", str(crec.redeemer_cbor_file)]
                )
            if crec.redeemer_value:
                grouped_args.extend(["--certificate-redeemer-value", str(crec.redeemer_value)])

        if crec.reference_txin:
            grouped_args.extend(
                [
                    "--certificate-tx-in-reference",
                    f"{crec.reference_txin.utxo_hash}#{crec.reference_txin.utxo_ix}",
                ]
            )

            crec_reference_type = crec.reference_type or consts.ScriptTypes.PLUTUS_V2
            if crec_reference_type == consts.ScriptTypes.PLUTUS_V2:
                grouped_args.append("--certificate-plutus-script-v2")
            if crec_reference_type == consts.ScriptTypes.PLUTUS_V3:
                grouped_args.append("--certificate-plutus-script-v3")

            if with_execution_units and crec.execution_units:
                grouped_args.extend(
                    [
                        "--certificate-reference-tx-in-execution-units",
                        f"({crec.execution_units[0]},{crec.execution_units[1]})",
                    ]
                )

            if crec.redeemer_file:
                grouped_args.extend(
                    ["--certificate-reference-tx-in-redeemer-file", str(crec.redeemer_file)]
                )
            if crec.redeemer_cbor_file:
                grouped_args.extend(
                    [
                        "--certificate-reference-tx-in-redeemer-cbor-file",
                        str(crec.redeemer_cbor_file),
                    ]
                )
            if crec.redeemer_value:
                grouped_args.extend(
                    ["--certificate-reference-tx-in-redeemer-value", str(crec.redeemer_value)]
                )

    # Proposals
    for prec in complex_proposals:
        prec_collaterals = {f"{c.utxo_hash}#{c.utxo_ix}" for c in prec.collaterals}
        collaterals_all.update(prec_collaterals)
        grouped_args.extend(
            [
                "--proposal-file",
                str(prec.proposal_file),
            ]
        )

        if prec.script_file:
            grouped_args.extend(
                [
                    "--proposal-script-file",
                    str(prec.script_file),
                ]
            )

            if with_execution_units and prec.execution_units:
                grouped_args.extend(
                    [
                        "--proposal-execution-units",
                        f"({prec.execution_units[0]},{prec.execution_units[1]})",
                    ]
                )

            if prec.redeemer_file:
                grouped_args.extend(["--proposal-redeemer-file", str(prec.redeemer_file)])
            if prec.redeemer_cbor_file:
                grouped_args.extend(["--proposal-redeemer-cbor-file", str(prec.redeemer_cbor_file)])
            if prec.redeemer_value:
                grouped_args.extend(["--proposal-redeemer-value", str(prec.redeemer_value)])

    # Withdrawals
    for wrec in script_withdrawals:
        wrec_collaterals = {f"{c.utxo_hash}#{c.utxo_ix}" for c in wrec.collaterals}
        collaterals_all.update(wrec_collaterals)
        grouped_args.extend(
            [
                "--withdrawal",
                f"{wrec.txout.address}+{wrec.txout.amount}",
            ]
        )

        if wrec.script_file:
            grouped_args.extend(
                [
                    "--withdrawal-script-file",
                    str(wrec.script_file),
                ]
            )

            if with_execution_units and wrec.execution_units:
                grouped_args.extend(
                    [
                        "--withdrawal-execution-units",
                        f"({wrec.execution_units[0]},{wrec.execution_units[1]})",
                    ]
                )

            if wrec.redeemer_file:
                grouped_args.extend(["--withdrawal-redeemer-file", str(wrec.redeemer_file)])
            if wrec.redeemer_cbor_file:
                grouped_args.extend(
                    ["--withdrawal-redeemer-cbor-file", str(wrec.redeemer_cbor_file)]
                )
            if wrec.redeemer_value:
                grouped_args.extend(["--withdrawal-redeemer-value", str(wrec.redeemer_value)])

        if wrec.reference_txin:
            grouped_args.extend(
                [
                    "--withdrawal-tx-in-reference",
                    f"{wrec.reference_txin.utxo_hash}#{wrec.reference_txin.utxo_ix}",
                ]
            )

            wrec_reference_type = wrec.reference_type or consts.ScriptTypes.PLUTUS_V2
            if wrec_reference_type == consts.ScriptTypes.PLUTUS_V2:
                grouped_args.append("--withdrawal-plutus-script-v2")
            elif wrec_reference_type == consts.ScriptTypes.PLUTUS_V3:
                grouped_args.append("--withdrawal-plutus-script-v3")

            if with_execution_units and wrec.execution_units:
                grouped_args.extend(
                    [
                        "--withdrawal-reference-tx-in-execution-units",
                        f"({wrec.execution_units[0]},{wrec.execution_units[1]})",
                    ]
                )

            if wrec.redeemer_file:
                grouped_args.extend(
                    ["--withdrawal-reference-tx-in-redeemer-file", str(wrec.redeemer_file)]
                )
            if wrec.redeemer_cbor_file:
                grouped_args.extend(
                    [
                        "--withdrawal-reference-tx-in-redeemer-cbor-file",
                        str(wrec.redeemer_cbor_file),
                    ]
                )
            if wrec.redeemer_value:
                grouped_args.extend(
                    ["--withdrawal-reference-tx-in-redeemer-value", str(wrec.redeemer_value)]
                )

    # Voting
    for vrec in script_votes:
        vrec_collaterals = {f"{c.utxo_hash}#{c.utxo_ix}" for c in vrec.collaterals}
        collaterals_all.update(vrec_collaterals)
        grouped_args.extend(
            [
                "--vote-file",
                str(vrec.vote_file),
            ]
        )

        if vrec.script_file:
            grouped_args.extend(
                [
                    "--vote-script-file",
                    str(vrec.script_file),
                ]
            )

            if with_execution_units and vrec.execution_units:
                grouped_args.extend(
                    [
                        "--vote-execution-units",
                        f"({vrec.execution_units[0]},{vrec.execution_units[1]})",
                    ]
                )

            if vrec.redeemer_file:
                grouped_args.extend(["--vote-redeemer-file", str(vrec.redeemer_file)])
            if vrec.redeemer_cbor_file:
                grouped_args.extend(["--vote-redeemer-cbor-file", str(vrec.redeemer_cbor_file)])
            if vrec.redeemer_value:
                grouped_args.extend(["--vote-redeemer-value", str(vrec.redeemer_value)])

    # Add unique collaterals
    grouped_args.extend(
        [
            *helpers._prepend_flag("--tx-in-collateral", collaterals_all),
        ]
    )

    return grouped_args


def get_proposal_file_argname(era_in_use: str = "") -> str:
    """Return the name of the proposal file argument."""
    proposal_file_argname = (
        "--proposal-file"
        if (consts.Eras[era_in_use.upper()].value >= consts.Eras.CONWAY.value)
        else "--update-proposal-file"
    )
    return proposal_file_argname



"""Group of methods for working with stake pools."""

import logging
import pathlib as pl

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class StakePoolGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj

    def gen_pool_metadata_hash(self, pool_metadata_file: itp.FileType) -> str:
        """Generate the hash of pool metadata.

        Args:
            pool_metadata_file: A path to the pool metadata file.

        Returns:
            str: A metadata hash.
        """
        return (
            self._clusterlib_obj.cli(
                ["stake-pool", "metadata-hash", "--pool-metadata-file", str(pool_metadata_file)]
            )
            .stdout.rstrip()
            .decode("ascii")
        )

    def gen_pool_registration_cert(
        self,
        pool_data: structs.PoolData,
        vrf_vkey_file: itp.FileType,
        cold_vkey_file: itp.FileType,
        owner_stake_vkey_files: itp.FileTypeList,
        reward_account_vkey_file: itp.FileType | None = None,
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Generate a stake pool registration certificate.

        Args:
            pool_data: A `structs.PoolData` data container containing info about the stake pool.
            vrf_vkey_file: A path to node VRF vkey file.
            cold_vkey_file: A path to pool cold vkey file.
            owner_stake_vkey_files: A list of paths to pool owner stake vkey files.
            reward_account_vkey_file: A path to pool reward account vkey file (optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the generated certificate.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{pool_data.pool_name}_pool_reg.cert"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        metadata_cmd = []
        if pool_data.pool_metadata_url and pool_data.pool_metadata_hash:
            metadata_cmd = [
                "--metadata-url",
                str(pool_data.pool_metadata_url),
                "--metadata-hash",
                str(pool_data.pool_metadata_hash),
            ]

        relay_cmd = []
        if pool_data.pool_relay_dns:
            relay_cmd.extend(["--single-host-pool-relay", pool_data.pool_relay_dns])
        if pool_data.pool_relay_ipv4:
            relay_cmd.extend(["--pool-relay-ipv4", pool_data.pool_relay_ipv4])
        if pool_data.pool_relay_port:
            relay_cmd.extend(["--pool-relay-port", str(pool_data.pool_relay_port)])

        self._clusterlib_obj.cli(
            [
                "stake-pool",
                "registration-certificate",
                "--pool-pledge",
                str(pool_data.pool_pledge),
                "--pool-cost",
                str(pool_data.pool_cost),
                "--pool-margin",
                str(pool_data.pool_margin),
                "--vrf-verification-key-file",
                str(vrf_vkey_file),
                "--cold-verification-key-file",
                str(cold_vkey_file),
                "--pool-reward-account-verification-key-file",
                str(reward_account_vkey_file)
                if reward_account_vkey_file
                else str(next(iter(owner_stake_vkey_files))),
                *helpers._prepend_flag(
                    "--pool-owner-stake-verification-key-file", owner_stake_vkey_files
                ),
                *self._clusterlib_obj.magic_args,
                "--out-file",
                str(out_file),
                *metadata_cmd,
                *relay_cmd,
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def gen_pool_deregistration_cert(
        self,
        pool_name: str,
        cold_vkey_file: itp.FileType,
        epoch: int,
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Generate a stake pool deregistration certificate.

        Args:
            pool_name: A name of the stake pool.
            cold_vkey_file: A path to pool cold vkey file.
            epoch: An epoch where the pool will be deregistered.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the generated certificate.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{pool_name}_pool_dereg.cert"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                "stake-pool",
                "deregistration-certificate",
                "--cold-verification-key-file",
                str(cold_vkey_file),
                "--epoch",
                str(epoch),
                "--out-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def get_stake_pool_id(
        self,
        cold_vkey_file: itp.FileType | None = None,
        stake_pool_vkey: str = "",
    ) -> str:
        """Return pool ID from the offline key.

        Args:
            cold_vkey_file: A path to pool cold vkey file.
            stake_pool_vkey: A stake pool verification key (Bech32 or hex-encoded).

        Returns:
            str: A pool ID.
        """
        if stake_pool_vkey:
            key_args = ["--stake-pool-verification-key", str(stake_pool_vkey)]
        elif cold_vkey_file:
            key_args = ["--cold-verification-key-file", str(cold_vkey_file)]
        else:
            msg = "No key was specified."
            raise AssertionError(msg)

        pool_id = (
            self._clusterlib_obj.cli(["stake-pool", "id", *key_args]).stdout.strip().decode("ascii")
        )
        return pool_id

    def create_stake_pool(
        self,
        pool_data: structs.PoolData,
        pool_owners: list[structs.PoolUser],
        tx_name: str,
        destination_dir: itp.FileType = ".",
    ) -> structs.PoolCreationOutput:
        """Create and register a stake pool.

        Args:
            pool_data: A `structs.PoolData` data container containing info about the stake pool.
            pool_owners: A list of `structs.PoolUser` structures containing pool user addresses
                and keys.
            tx_name: A name of the transaction.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.PoolCreationOutput: A data container containing pool creation output.
        """
        # Create the KES key pair
        node_kes = self._clusterlib_obj.g_node.gen_kes_key_pair(
            node_name=pool_data.pool_name,
            destination_dir=destination_dir,
        )
        LOGGER.debug(f"KES keys created - {node_kes.vkey_file}; {node_kes.skey_file}")

        # Create the VRF key pair
        node_vrf = self._clusterlib_obj.g_node.gen_vrf_key_pair(
            node_name=pool_data.pool_name,
            destination_dir=destination_dir,
        )
        LOGGER.debug(f"VRF keys created - {node_vrf.vkey_file}; {node_vrf.skey_file}")

        # Create the cold key pair and node operational certificate counter
        node_cold = self._clusterlib_obj.g_node.gen_cold_key_pair_and_counter(
            node_name=pool_data.pool_name,
            destination_dir=destination_dir,
        )
        LOGGER.debug(
            "Cold keys created and counter created - "
            f"{node_cold.vkey_file}; {node_cold.skey_file}; {node_cold.counter_file}"
        )

        pool_reg_cert_file, tx_raw_output = self.register_stake_pool(
            pool_data=pool_data,
            pool_owners=pool_owners,
            vrf_vkey_file=node_vrf.vkey_file,
            cold_key_pair=node_cold,
            tx_name=tx_name,
            destination_dir=destination_dir,
        )

        return structs.PoolCreationOutput(
            stake_pool_id=self.get_stake_pool_id(node_cold.vkey_file),
            vrf_key_pair=node_vrf,
            cold_key_pair=node_cold,
            pool_reg_cert_file=pool_reg_cert_file,
            pool_data=pool_data,
            pool_owners=pool_owners,
            tx_raw_output=tx_raw_output,
            kes_key_pair=node_kes,
        )

    def register_stake_pool(
        self,
        pool_data: structs.PoolData,
        pool_owners: list[structs.PoolUser],
        vrf_vkey_file: itp.FileType,
        cold_key_pair: structs.ColdKeyPair,
        tx_name: str,
        reward_account_vkey_file: itp.FileType | None = None,
        deposit: int | None = None,
        destination_dir: itp.FileType = ".",
    ) -> tuple[pl.Path, structs.TxRawOutput]:
        """Register a stake pool.

        Args:
            pool_data: A `structs.PoolData` data container containing info about the stake pool.
            pool_owners: A list of `structs.PoolUser` structures containing pool user addresses
                and keys.
            vrf_vkey_file: A path to node VRF vkey file.
            cold_key_pair: A `structs.ColdKeyPair` data container containing the key pair
                and the counter.
            tx_name: A name of the transaction.
            reward_account_vkey_file: A path to reward account vkey file (optional).
            deposit: A deposit amount needed by the transaction (optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Tuple[Path, structs.TxRawOutput]: A tuple with pool registration cert file and
                transaction output details.
        """
        tx_name = f"{tx_name}_reg_pool"
        pool_reg_cert_file = self.gen_pool_registration_cert(
            pool_data=pool_data,
            vrf_vkey_file=vrf_vkey_file,
            cold_vkey_file=cold_key_pair.vkey_file,
            owner_stake_vkey_files=[p.stake.vkey_file for p in pool_owners],
            reward_account_vkey_file=reward_account_vkey_file,
            destination_dir=destination_dir,
        )

        # Submit the pool registration certificate through a tx
        tx_files = structs.TxFiles(
            certificate_files=[pool_reg_cert_file],
            signing_key_files=[
                *[p.payment.skey_file for p in pool_owners],
                *[p.stake.skey_file for p in pool_owners],
                cold_key_pair.skey_file,
            ],
        )

        tx_raw_output = self._clusterlib_obj.g_transaction.send_tx(
            src_address=pool_owners[0].payment.address,
            tx_name=tx_name,
            tx_files=tx_files,
            deposit=deposit,
            destination_dir=destination_dir,
        )

        return pool_reg_cert_file, tx_raw_output

    def deregister_stake_pool(
        self,
        pool_owners: list[structs.PoolUser],
        cold_key_pair: structs.ColdKeyPair,
        epoch: int,
        pool_name: str,
        tx_name: str,
        destination_dir: itp.FileType = ".",
    ) -> tuple[pl.Path, structs.TxRawOutput]:
        """Deregister a stake pool.

        Args:
            pool_owners: A list of `structs.PoolUser` structures containing pool user addresses
                and keys.
            cold_key_pair: A `structs.ColdKeyPair` data container containing the key pair
                and the counter.
            epoch: An epoch where the update proposal will take effect (optional).
            pool_name: A name of the stake pool.
            tx_name: A name of the transaction.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Tuple[Path, structs.TxRawOutput]: A data container with pool registration cert file and
                transaction output details.
        """
        tx_name = f"{tx_name}_dereg_pool"
        LOGGER.debug(
            f"Deregistering stake pool starting with epoch: {epoch}; "
            f"Current epoch is: {self._clusterlib_obj.g_query.get_epoch()}"
        )
        pool_dereg_cert_file = self.gen_pool_deregistration_cert(
            pool_name=pool_name,
            cold_vkey_file=cold_key_pair.vkey_file,
            epoch=epoch,
            destination_dir=destination_dir,
        )

        # Submit the pool deregistration certificate through a tx
        tx_files = structs.TxFiles(
            certificate_files=[pool_dereg_cert_file],
            signing_key_files=[
                *[p.payment.skey_file for p in pool_owners],
                *[p.stake.skey_file for p in pool_owners],
                cold_key_pair.skey_file,
            ],
        )

        tx_raw_output = self._clusterlib_obj.g_transaction.send_tx(
            src_address=pool_owners[0].payment.address,
            tx_name=tx_name,
            tx_files=tx_files,
            destination_dir=destination_dir,
        )

        return pool_dereg_cert_file, tx_raw_output

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: clusterlib_obj={id(self._clusterlib_obj)}>"



"""Group of methods for Conway governance committee commands."""

import logging
import pathlib as pl

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class GovCommitteeGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj
        self._group_args = ("governance", "committee")

    def _get_cold_vkey_args(
        self,
        cold_vkey: str = "",
        cold_vkey_file: itp.FileType = "",
        cold_vkey_hash: str = "",
    ) -> list[str]:
        """Get arguments for cold verification key."""
        if cold_vkey:
            key_args = ["--cold-verification-key", str(cold_vkey)]
        elif cold_vkey_file:
            key_args = ["--cold-verification-key-file", str(cold_vkey_file)]
        elif cold_vkey_hash:
            key_args = ["--cold-key-hash", str(cold_vkey_hash)]
        else:
            msg = "Either `cold_vkey`, `cold_vkey_file` or `cold_vkey_hash` is needed."
            raise AssertionError(msg)

        return key_args

    def gen_cold_key_resignation_cert(
        self,
        key_name: str,
        cold_vkey: str = "",
        cold_vkey_file: itp.FileType = "",
        cold_vkey_hash: str = "",
        resignation_metadata_url: str = "",
        resignation_metadata_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Create cold key resignation certificate for a Constitutional Committee Member."""
        destination_dir = pl.Path(destination_dir).expanduser()
        cert_file = destination_dir / f"{key_name}_committee_cold_resignation.cert"
        clusterlib_helpers._check_files_exist(cert_file, clusterlib_obj=self._clusterlib_obj)

        key_args = self._get_cold_vkey_args(
            cold_vkey=cold_vkey, cold_vkey_file=cold_vkey_file, cold_vkey_hash=cold_vkey_hash
        )

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create-cold-key-resignation-certificate",
                *key_args,
                "--resignation-metadata-url",
                resignation_metadata_url,
                "--resignation-metadata-hash",
                resignation_metadata_hash,
                "--out-file",
                str(cert_file),
            ]
        )

        helpers._check_outfiles(cert_file)
        return cert_file

    def gen_hot_key_auth_cert(
        self,
        key_name: str,
        cold_vkey: str = "",
        cold_vkey_file: itp.FileType = "",
        cold_vkey_hash: str = "",
        hot_key: str = "",
        hot_key_file: itp.FileType = "",
        hot_key_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Create hot key authorization certificate for a Constitutional Committee Member."""
        destination_dir = pl.Path(destination_dir).expanduser()
        cert_file = destination_dir / f"{key_name}_committee_hot_auth.cert"
        clusterlib_helpers._check_files_exist(cert_file, clusterlib_obj=self._clusterlib_obj)

        cold_vkey_args = self._get_cold_vkey_args(
            cold_vkey=cold_vkey, cold_vkey_file=cold_vkey_file, cold_vkey_hash=cold_vkey_hash
        )

        if hot_key:
            hot_key_args = ["--hot-verification-key", str(hot_key)]
        elif hot_key_file:
            hot_key_args = ["--hot-verification-key-file", str(hot_key_file)]
        elif hot_key_hash:
            hot_key_args = ["--hot-verification-key-hash", str(hot_key_hash)]
        else:
            msg = "Either `hot_key`, `hot_key_file` or `hot_key_hash` is needed."
            raise AssertionError(msg)

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create-hot-key-authorization-certificate",
                *cold_vkey_args,
                *hot_key_args,
                "--out-file",
                str(cert_file),
            ]
        )

        helpers._check_outfiles(cert_file)
        return cert_file

    def gen_cold_key_pair(
        self, key_name: str, destination_dir: itp.FileType = "."
    ) -> structs.KeyPair:
        """Create a cold key pair for a Constitutional Committee Member."""
        destination_dir = pl.Path(destination_dir).expanduser()
        vkey = destination_dir / f"{key_name}_committee_cold.vkey"
        skey = destination_dir / f"{key_name}_committee_cold.skey"
        clusterlib_helpers._check_files_exist(vkey, skey, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "key-gen-cold",
                "--cold-verification-key-file",
                str(vkey),
                "--cold-signing-key-file",
                str(skey),
            ]
        )

        helpers._check_outfiles(vkey, skey)
        return structs.KeyPair(vkey, skey)

    def gen_hot_key_pair(
        self, key_name: str, destination_dir: itp.FileType = "."
    ) -> structs.KeyPair:
        """Create a cold key pair for a Constitutional Committee Member."""
        destination_dir = pl.Path(destination_dir).expanduser()
        vkey = destination_dir / f"{key_name}_committee_hot.vkey"
        skey = destination_dir / f"{key_name}_committee_hot.skey"
        clusterlib_helpers._check_files_exist(vkey, skey, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "key-gen-hot",
                "--verification-key-file",
                str(vkey),
                "--signing-key-file",
                str(skey),
            ]
        )

        helpers._check_outfiles(vkey, skey)
        return structs.KeyPair(vkey, skey)

    def get_key_hash(
        self,
        vkey: str = "",
        vkey_file: itp.FileType = "",
    ) -> str:
        """Get the identifier (hash) of a public key."""
        vkey_file = pl.Path(vkey_file).expanduser()
        clusterlib_helpers._check_files_exist(vkey_file, clusterlib_obj=self._clusterlib_obj)

        if vkey:
            key_args = ["--verification-key", str(vkey)]
        elif vkey_file:
            key_args = ["--verification-key-file", str(vkey_file)]
        else:
            msg = "Either `vkey` or `vkey_file` is needed."
            raise AssertionError(msg)

        key_hash = (
            self._clusterlib_obj.cli([*self._group_args, "key-hash", *key_args])
            .stdout.rstrip()
            .decode("ascii")
        )

        return key_hash



import pathlib as pl
import typing as tp

if tp.TYPE_CHECKING:
    from cardano_clusterlib.clusterlib_klass import ClusterLib  # noqa: F401

FileType = str | pl.Path
FileTypeList = list[FileType] | list[str] | list[pl.Path]
# List of `FileType`s, empty list, or empty tuple
OptionalFiles = FileTypeList | tuple[()]
# TODO: needed until https://github.com/python/typing/issues/256 is fixed
UnpackableSequence = list | tuple | set | frozenset



"""Wrapper for cardano-cli for working with cardano cluster."""

import json
import logging
import pathlib as pl
import subprocess
import time

from packaging import version

from cardano_clusterlib import address_group
from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import consts
from cardano_clusterlib import exceptions
from cardano_clusterlib import genesis_group
from cardano_clusterlib import gov_group
from cardano_clusterlib import helpers
from cardano_clusterlib import key_group
from cardano_clusterlib import legacy_gov_group
from cardano_clusterlib import node_group
from cardano_clusterlib import query_group
from cardano_clusterlib import stake_address_group
from cardano_clusterlib import stake_pool_group
from cardano_clusterlib import structs
from cardano_clusterlib import transaction_group
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class ClusterLib:
    """Methods for working with cardano cluster using `cardano-cli`..

    Attributes:
        state_dir: A directory with cluster state files (keys, config files, logs, ...).
        protocol: A cluster protocol - full cardano mode by default.
        slots_offset: Difference in slots between cluster's start era and Shelley era
            (Byron vs Shelley)
        socket_path: A path to socket file for communication with the node. This overrides the
            `CARDANO_NODE_SOCKET_PATH` environment variable.
        command_era: An era used for CLI commands, by default same as the latest network Era.
    """

    def __init__(
        self,
        state_dir: itp.FileType,
        slots_offset: int | None = None,
        socket_path: itp.FileType = "",
        command_era: str = consts.CommandEras.LATEST,
    ):
        try:
            self.command_era = getattr(consts.CommandEras, command_era.upper())
        except AttributeError as excp:
            msg = f"Unknown command era `{command_era}`."
            raise exceptions.CLIError(msg) from excp

        self.cluster_id = 0  # Can be used for identifying cluster instance
        # Number of new blocks before the Tx is considered confirmed
        self.confirm_blocks = consts.CONFIRM_BLOCKS_NUM
        self._rand_str = helpers.get_rand_str(4)
        self._cli_log = ""
        self.era_in_use = (
            consts.Eras.__members__.get(command_era.upper()) or consts.Eras["DEFAULT"]
        ).name.lower()

        self.state_dir = pl.Path(state_dir).expanduser().resolve()
        if not self.state_dir.exists():
            msg = f"The state dir `{self.state_dir}` doesn't exist."
            raise exceptions.CLIError(msg)

        self._init_socket_path = socket_path
        self.socket_path: pl.Path | None = None
        self.socket_args: list[str] = []
        self.set_socket_path(socket_path=socket_path)

        self.pparams_file = self.state_dir / f"pparams-{self._rand_str}.json"

        self.genesis_json = clusterlib_helpers._find_genesis_json(clusterlib_obj=self)
        with open(self.genesis_json, encoding="utf-8") as in_json:
            self.genesis = json.load(in_json)

        self.slot_length = self.genesis["slotLength"]
        self.epoch_length = self.genesis["epochLength"]
        self.epoch_length_sec = self.epoch_length * self.slot_length
        self.slots_per_kes_period = self.genesis["slotsPerKESPeriod"]
        self.max_kes_evolutions = self.genesis["maxKESEvolutions"]

        self.network_magic = self.genesis["networkMagic"]
        if self.network_magic == consts.MAINNET_MAGIC:
            self.magic_args = ["--mainnet"]
        else:
            self.magic_args = ["--testnet-magic", str(self.network_magic)]

        self._slots_offset = slots_offset if slots_offset is not None else None

        self.ttl_length = 1000
        # TODO: proper calculation based on `utxoCostPerWord` needed
        self._min_change_value = 1800_000

        # Conway+ era
        self.conway_genesis_json: pl.Path | None = None
        self.conway_genesis: dict = {}
        if consts.Eras[self.era_in_use.upper()].value >= consts.Eras.CONWAY.value:
            # Conway genesis
            self.conway_genesis_json = clusterlib_helpers._find_conway_genesis_json(
                clusterlib_obj=self
            )
            with open(self.conway_genesis_json, encoding="utf-8") as in_json:
                self.conway_genesis = json.load(in_json)

        self.overwrite_outfiles = True

        self._cli_version: version.Version | None = None

        # Groups of commands
        self._transaction_group: transaction_group.TransactionGroup | None = None
        self._query_group: query_group.QueryGroup | None = None
        self._address_group: address_group.AddressGroup | None = None
        self._stake_address_group: stake_address_group.StakeAddressGroup | None = None
        self._stake_pool_group: stake_pool_group.StakePoolGroup | None = None
        self._node_group: node_group.NodeGroup | None = None
        self._key_group: key_group.KeyGroup | None = None
        self._genesis_group: genesis_group.GenesisGroup | None = None
        self._legacy_gov_group: legacy_gov_group.LegacyGovGroup | None = None
        self._governance_group: gov_group.GovernanceGroup | None = None

    def set_socket_path(self, socket_path: itp.FileType | None) -> None:
        """Set a path to socket file for communication with the node."""
        if not socket_path:
            self.socket_path = None
            self.socket_args = []
            return

        socket_path = pl.Path(socket_path).expanduser().resolve()
        if not socket_path.exists():
            msg = f"The socket `{socket_path}` doesn't exist."
            raise exceptions.CLIError(msg)

        self.socket_path = socket_path
        self.socket_args = ["--socket-path", str(self.socket_path)]

    @property
    def cli_version(self) -> version.Version:
        """Version of `cardano-cli`."""
        if self._cli_version is None:
            version_out = self.cli(
                ["cardano-cli", "--version"], add_default_args=False
            ).stdout.decode()
            version_str = version_out.split(" ")[1]
            self._cli_version = version.parse(version_str)
        return self._cli_version

    @property
    def slots_offset(self) -> int:
        """Get offset of slots from Byron era vs current configuration."""
        if self._slots_offset is None:
            self._slots_offset = clusterlib_helpers.get_slots_offset(clusterlib_obj=self)
        return self._slots_offset

    @property
    def g_transaction(self) -> transaction_group.TransactionGroup:
        """Transaction group."""
        if not self._transaction_group:
            self._transaction_group = transaction_group.TransactionGroup(clusterlib_obj=self)
        return self._transaction_group

    @property
    def g_query(self) -> query_group.QueryGroup:
        """Query group."""
        if not self._query_group:
            self._query_group = query_group.QueryGroup(clusterlib_obj=self)
        return self._query_group

    @property
    def g_address(self) -> address_group.AddressGroup:
        """Address group."""
        if not self._address_group:
            self._address_group = address_group.AddressGroup(clusterlib_obj=self)
        return self._address_group

    @property
    def g_stake_address(self) -> stake_address_group.StakeAddressGroup:
        """Stake address group."""
        if not self._stake_address_group:
            self._stake_address_group = stake_address_group.StakeAddressGroup(clusterlib_obj=self)
        return self._stake_address_group

    @property
    def g_stake_pool(self) -> stake_pool_group.StakePoolGroup:
        """Stake pool group."""
        if not self._stake_pool_group:
            self._stake_pool_group = stake_pool_group.StakePoolGroup(clusterlib_obj=self)
        return self._stake_pool_group

    @property
    def g_node(self) -> node_group.NodeGroup:
        """Node group."""
        if not self._node_group:
            self._node_group = node_group.NodeGroup(clusterlib_obj=self)
        return self._node_group

    @property
    def g_key(self) -> key_group.KeyGroup:
        """Key group."""
        if not self._key_group:
            self._key_group = key_group.KeyGroup(clusterlib_obj=self)
        return self._key_group

    @property
    def g_genesis(self) -> genesis_group.GenesisGroup:
        """Genesis group."""
        if not self._genesis_group:
            self._genesis_group = genesis_group.GenesisGroup(clusterlib_obj=self)
        return self._genesis_group

    @property
    def g_legacy_governance(self) -> legacy_gov_group.LegacyGovGroup:
        """Legacy governance group."""
        if not self._legacy_gov_group:
            self._legacy_gov_group = legacy_gov_group.LegacyGovGroup(clusterlib_obj=self)
        return self._legacy_gov_group

    @property
    def g_governance(self) -> gov_group.GovernanceGroup:
        """Governance group."""
        if self._governance_group:
            return self._governance_group

        if not self.conway_genesis:
            msg = "The governance group can be used only with Command era >= Conway."
            raise exceptions.CLIError(msg)

        self._governance_group = gov_group.GovernanceGroup(clusterlib_obj=self)
        return self._governance_group

    def cli(
        self,
        cli_args: list[str],
        timeout: float | None = None,
        add_default_args: bool = True,
    ) -> structs.CLIOut:
        """Run the `cardano-cli` command.

        Args:
            cli_args: A list of arguments for cardano-cli.
            timeout: A timeout for the command, in seconds (optional).
            add_default_args: Whether to add default arguments to the command (optional).

        Returns:
            structs.CLIOut: A data container containing command stdout and stderr.
        """
        cli_args_strs_all = [str(arg) for arg in cli_args]

        if add_default_args:
            cli_args_strs_all.insert(0, "cardano-cli")
            cli_args_strs_all.insert(1, self.command_era)

        cli_args_strs = [arg for arg in cli_args_strs_all if arg != consts.SUBCOMMAND_MARK]

        cmd_str = clusterlib_helpers._format_cli_args(cli_args=cli_args_strs)
        clusterlib_helpers._write_cli_log(clusterlib_obj=self, command=cmd_str)
        LOGGER.debug("Running `%s`", cmd_str)

        # Re-run the command when running into
        # Network.Socket.connect: <socket: X>: resource exhausted (Resource temporarily unavailable)
        # or
        # MuxError (MuxIOException writev: resource vanished (Broken pipe)) "(sendAll errored)"
        for __ in range(3):
            retcode = None
            with subprocess.Popen(
                cli_args_strs, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ) as p:
                stdout, stderr = p.communicate(timeout=timeout)
                retcode = p.returncode

            if retcode == 0:
                break

            stderr_dec = stderr.decode()
            err_msg = (
                f"An error occurred running a CLI command `{cmd_str}` on path "
                f"`{pl.Path.cwd()}`: {stderr_dec}"
            )
            if "resource exhausted" in stderr_dec or "resource vanished" in stderr_dec:
                LOGGER.error(err_msg)
                time.sleep(0.4)
                continue
            raise exceptions.CLIError(err_msg)
        else:
            raise exceptions.CLIError(err_msg)

        return structs.CLIOut(stdout or b"", stderr or b"")

    def refresh_pparams_file(self) -> None:
        """Refresh protocol parameters file."""
        self.g_query.query_cli(["protocol-parameters", "--out-file", str(self.pparams_file)])

    def create_pparams_file(self) -> None:
        """Create protocol parameters file if it doesn't exist."""
        if self.pparams_file.exists():
            return
        self.refresh_pparams_file()

    def wait_for_new_block(self, new_blocks: int = 1) -> int:
        """Wait for new block(s) to be created.

        Args:
            new_blocks: A number of new blocks to wait for (optional).

        Returns:
            int: A block number of last added block.
        """
        initial_tip = self.g_query.get_tip()
        initial_block = int(initial_tip["block"])

        if new_blocks < 1:
            return initial_block

        return clusterlib_helpers.wait_for_block(
            clusterlib_obj=self, tip=initial_tip, block_no=initial_block + new_blocks
        )

    def wait_for_block(self, block: int) -> int:
        """Wait for block number.

        Args:
            block: A block number to wait for.

        Returns:
            int: A block number of last added block.
        """
        return clusterlib_helpers.wait_for_block(
            clusterlib_obj=self, tip=self.g_query.get_tip(), block_no=block
        )

    def wait_for_slot(self, slot: int) -> int:
        """Wait for slot number.

        Args:
            slot: A slot number to wait for.

        Returns:
            int: A slot number of last block.
        """
        min_sleep = 1.5  # in sec
        long_sleep = 15  # in sec
        no_block_time = 0  # in slots
        next_block_timeout = 300  # in slots
        last_slot = -1
        printed = False
        for __ in range(100):
            this_slot = self.g_query.get_slot_no()

            slots_diff = slot - this_slot
            if slots_diff <= 0:
                return this_slot

            if this_slot == last_slot:
                if no_block_time >= next_block_timeout:
                    msg = f"Failed to wait for slot number {slot}, no new blocks are being created."
                    raise exceptions.CLIError(msg)
            else:
                no_block_time = 0

            _sleep_time = slots_diff * self.slot_length
            sleep_time = max(min_sleep, _sleep_time)

            if not printed and sleep_time > long_sleep:
                LOGGER.info(f"Waiting for {sleep_time:.2f} sec for slot no {slot}.")
                printed = True

            last_slot = this_slot
            no_block_time += slots_diff
            time.sleep(sleep_time)

        msg = f"Failed to wait for slot number {slot}."
        raise exceptions.CLIError(msg)

    def wait_for_new_epoch(self, new_epochs: int = 1, padding_seconds: int = 0) -> int:
        """Wait for new epoch(s).

        Args:
            new_epochs: A number of new epochs to wait for (optional).
            padding_seconds: A number of additional seconds to wait for (optional).

        Returns:
            int: The current epoch.
        """
        start_tip = self.g_query.get_tip()
        start_epoch = int(start_tip["epoch"])

        if new_epochs < 1:
            return start_epoch

        epoch_no = start_epoch + new_epochs
        return clusterlib_helpers.wait_for_epoch(
            clusterlib_obj=self, tip=start_tip, epoch_no=epoch_no, padding_seconds=padding_seconds
        )

    def wait_for_epoch(
        self, epoch_no: int, padding_seconds: int = 0, future_is_ok: bool = True
    ) -> int:
        """Wait for epoch no.

        Args:
            epoch_no: A number of epoch to wait for.
            padding_seconds: A number of additional seconds to wait for (optional).
            future_is_ok: A bool indicating whether current epoch > `epoch_no` is acceptable
                (default: True).

        Returns:
            int: The current epoch.
        """
        return clusterlib_helpers.wait_for_epoch(
            clusterlib_obj=self,
            tip=self.g_query.get_tip(),
            epoch_no=epoch_no,
            padding_seconds=padding_seconds,
            future_is_ok=future_is_ok,
        )

    def time_to_epoch_end(self, tip: dict | None = None) -> float:
        """How many seconds to go to start of a new epoch."""
        tip = tip or self.g_query.get_tip()
        epoch = int(tip["epoch"])
        slot = int(tip["slot"])
        slots_to_go = (epoch + 1) * self.epoch_length - (slot + self.slots_offset - 1)
        return float(slots_to_go * self.slot_length)

    def time_from_epoch_start(self, tip: dict | None = None) -> float:
        """How many seconds passed from start of the current epoch."""
        s_to_epoch_stop = self.time_to_epoch_end(tip=tip)
        return float(self.epoch_length_sec - s_to_epoch_stop)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: command_era={self.command_era}>"



"""Group of methods for Conway governance DRep commands."""

import logging
import pathlib as pl

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import exceptions
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class GovDrepGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj
        self._group_args = ("governance", "drep")
        self._has_output_hex_prop: bool | None = None

    @property
    def _has_output_hex(self) -> bool:
        """Check if `drep id` has a `--output-hex` option."""
        if self._has_output_hex_prop is not None:
            return self._has_output_hex_prop

        err = ""
        try:
            self._clusterlib_obj.cli(
                ["cardano-cli", "conway", "governance", "drep", "id", "--output-hex"],
                add_default_args=False,
            )
        except exceptions.CLIError as excp:
            err = str(excp)

        self._has_output_hex_prop = "Invalid option" not in err
        return self._has_output_hex_prop

    def _get_cred_args(
        self,
        drep_script_hash: str = "",
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        drep_key_hash: str = "",
    ) -> list[str]:
        """Get arguments for script or vkey credentials."""
        if drep_script_hash:
            cred_args = ["--drep-script-hash", str(drep_script_hash)]
        elif drep_vkey:
            cred_args = ["--drep-verification-key", str(drep_vkey)]
        elif drep_vkey_file:
            cred_args = ["--drep-verification-key-file", str(drep_vkey_file)]
        elif drep_key_hash:
            cred_args = ["--drep-key-hash", str(drep_key_hash)]
        else:
            msg = (
                "Either `script_hash`, `drep_vkey`, `drep_vkey_file` or `drep_key_hash` is needed."
            )
            raise AssertionError(msg)

        return cred_args

    def gen_key_pair(self, key_name: str, destination_dir: itp.FileType = ".") -> structs.KeyPair:
        """Generate DRep verification and signing keys.

        Args:
            key_name: A name of the key pair.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.KeyPair: A data container containing the key pair.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        vkey = destination_dir / f"{key_name}_drep.vkey"
        skey = destination_dir / f"{key_name}_drep.skey"
        clusterlib_helpers._check_files_exist(vkey, skey, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "key-gen",
                "--verification-key-file",
                str(vkey),
                "--signing-key-file",
                str(skey),
            ]
        )

        helpers._check_outfiles(vkey, skey)
        return structs.KeyPair(vkey, skey)

    def get_id(
        self,
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        out_format: str = "",
    ) -> str:
        """Return a DRep id.

        Args:
            drep_vkey: DRep verification key (Bech32 or hex-encoded, optional).
            drep_vkey_file: A path to corresponding drep vkey file (optional).
            out_format: Output format (optional, bech32 by default).

        Returns:
            str: A generated DRep id.
        """
        if drep_vkey:
            cli_args = ["--drep-verification-key", str(drep_vkey)]
        elif drep_vkey_file:
            cli_args = ["--drep-verification-key-file", str(drep_vkey_file)]
        else:
            msg = "Either `drep_vkey` or `drep_vkey_file` is needed."
            raise AssertionError(msg)

        if out_format:
            if out_format not in ("hex", "bech32"):
                msg = f"Invalid output format: {out_format} (expected 'hex' or 'bech32')."
                raise AssertionError(msg)
            if self._has_output_hex:
                cli_args.append(f"--output-{out_format}")
            else:
                cli_args.extend(["--output-format", str(out_format)])

        drep_id = (
            self._clusterlib_obj.cli(
                [
                    *self._group_args,
                    "id",
                    *cli_args,
                ]
            )
            .stdout.strip()
            .decode("ascii")
        )

        return drep_id

    def gen_registration_cert(
        self,
        cert_name: str,
        deposit_amt: int,
        drep_script_hash: str = "",
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        drep_key_hash: str = "",
        drep_metadata_url: str = "",
        drep_metadata_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Generate a DRep registration certificate.

        Args:
            cert_name: A name of the cert.
            deposit_amt: A key registration deposit amount.
            drep_script_hash: DRep script hash (hex-encoded, optional).
            drep_vkey: DRep verification key (Bech32 or hex-encoded, optional).
            drep_vkey_file: Filepath of the DRep verification key (optional).
            drep_key_hash: DRep verification key hash
                (either Bech32-encoded or hex-encoded, optional).
            drep_metadata_url: URL to the metadata file (optional).
            drep_metadata_hash: Hash of the metadata file (optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the generated certificate.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{cert_name}_drep_reg.cert"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        cred_args = self._get_cred_args(
            drep_script_hash=drep_script_hash,
            drep_vkey=drep_vkey,
            drep_vkey_file=drep_vkey_file,
            drep_key_hash=drep_key_hash,
        )

        metadata_args = []
        if drep_metadata_url:
            metadata_args = [
                "--drep-metadata-url",
                str(drep_metadata_url),
                "--drep-metadata-hash",
                str(drep_metadata_hash),
            ]

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "registration-certificate",
                *cred_args,
                "--key-reg-deposit-amt",
                str(deposit_amt),
                *metadata_args,
                "--out-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def gen_update_cert(
        self,
        cert_name: str,
        deposit_amt: int,
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        drep_key_hash: str = "",
        drep_metadata_url: str = "",
        drep_metadata_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Generate a DRep update certificate.

        Args:
            cert_name: A name of the cert.
            deposit_amt: A key registration deposit amount.
            drep_vkey: DRep verification key (Bech32 or hex-encoded, optional).
            drep_vkey_file: Filepath of the DRep verification key (optional).
            drep_key_hash: DRep verification key hash
                (either Bech32-encoded or hex-encoded, optional).
            drep_metadata_url: URL to the metadata file (optional).
            drep_metadata_hash: Hash of the metadata file (optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the generated certificate.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{cert_name}_drep_update.cert"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        cred_args = self._get_cred_args(
            drep_vkey=drep_vkey,
            drep_vkey_file=drep_vkey_file,
            drep_key_hash=drep_key_hash,
        )

        metadata_args = []
        if drep_metadata_url:
            metadata_args = [
                "--drep-metadata-url",
                str(drep_metadata_url),
                "--drep-metadata-hash",
                str(drep_metadata_hash),
            ]

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "update-certificate",
                *cred_args,
                "--key-reg-deposit-amt",
                str(deposit_amt),
                *metadata_args,
                "--out-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def gen_retirement_cert(
        self,
        cert_name: str,
        deposit_amt: int,
        drep_script_hash: str = "",
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        drep_key_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Generate a DRep retirement certificate.

        Args:
            cert_name: A name of the cert.
            deposit_amt: A key registration deposit amount.
            drep_script_hash: DRep script hash (hex-encoded, optional).
            drep_vkey: DRep verification key (Bech32 or hex-encoded, optional).
            drep_vkey_file: Filepath of the DRep verification key (optional).
            drep_key_hash: DRep verification key hash
                (either Bech32-encoded or hex-encoded, optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the generated certificate.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{cert_name}_drep_retirement.cert"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        cred_args = self._get_cred_args(
            drep_script_hash=drep_script_hash,
            drep_vkey=drep_vkey,
            drep_vkey_file=drep_vkey_file,
            drep_key_hash=drep_key_hash,
        )

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "retirement-certificate",
                *cred_args,
                "--deposit-amt",
                str(deposit_amt),
                "--out-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def get_metadata_hash(
        self,
        drep_metadata_file: itp.FileType,
    ) -> str:
        """Get the hash of the metadata.

        Args:
            drep_metadata_file: A path to the metadata file.

        Returns:
            str: A hash of the metadata.
        """
        drep_metadata_file = pl.Path(drep_metadata_file).expanduser()
        clusterlib_helpers._check_files_exist(
            drep_metadata_file, clusterlib_obj=self._clusterlib_obj
        )

        metadata_hash = (
            self._clusterlib_obj.cli(
                [
                    *self._group_args,
                    "metadata-hash",
                    "--drep-metadata-file",
                    str(drep_metadata_file),
                ]
            )
            .stdout.rstrip()
            .decode("ascii")
        )

        return metadata_hash



"""Group of methods for working with payment addresses."""

import json
import logging
import pathlib as pl

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class AddressGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj

    def gen_payment_addr(
        self,
        addr_name: str,
        payment_vkey: str | None = None,
        payment_vkey_file: itp.FileType | None = None,
        payment_script_file: itp.FileType | None = None,
        stake_vkey: str | None = None,
        stake_vkey_file: itp.FileType | None = None,
        stake_script_file: itp.FileType | None = None,
        stake_address: str | None = None,
        destination_dir: itp.FileType = ".",
    ) -> str:
        """Generate a payment address, with optional delegation to a stake address.

        Args:
            addr_name: A name of payment address.
            payment_vkey: A vkey file (Bech32, optional).
            payment_vkey_file: A path to corresponding vkey file (optional).
            payment_script_file: A path to corresponding payment script file (optional).
            stake_vkey: A stake vkey file (optional).
            stake_vkey_file: A path to corresponding stake vkey file (optional).
            stake_script_file: A path to corresponding stake script file (optional).
            stake_address: A stake address (Bech32, optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            str: A generated payment address.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{addr_name}.addr"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        if payment_vkey_file:
            cli_args = ["--payment-verification-key-file", str(payment_vkey_file)]
        elif payment_script_file:
            cli_args = ["--payment-script-file", str(payment_script_file)]
        elif payment_vkey:
            cli_args = ["--payment-verification-key", str(payment_vkey)]
        else:
            msg = "Either `payment_vkey_file`, `payment_script_file` or `payment_vkey` is needed."
            raise AssertionError(msg)

        if stake_vkey:
            cli_args.extend(["--stake-verification-key", str(stake_vkey)])
        elif stake_vkey_file:
            cli_args.extend(["--stake-verification-key-file", str(stake_vkey_file)])
        elif stake_script_file:
            cli_args.extend(["--stake-script-file", str(stake_script_file)])
        elif stake_address:
            cli_args.extend(["--stake-address", str(stake_address)])

        self._clusterlib_obj.cli(
            [
                "address",
                "build",
                *self._clusterlib_obj.magic_args,
                *cli_args,
                "--out-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return helpers.read_address_from_file(out_file)

    def gen_payment_key_pair(
        self, key_name: str, extended: bool = False, destination_dir: itp.FileType = "."
    ) -> structs.KeyPair:
        """Generate an address key pair.

        Args:
            key_name: A name of the key pair.
            extended: A bool indicating whether to generate extended ed25519 Shelley-era key
                (False by default).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.KeyPair: A data container containing the key pair.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        vkey = destination_dir / f"{key_name}.vkey"
        skey = destination_dir / f"{key_name}.skey"
        clusterlib_helpers._check_files_exist(vkey, skey, clusterlib_obj=self._clusterlib_obj)

        extended_args = ["--extended-key"] if extended else []

        self._clusterlib_obj.cli(
            [
                "address",
                "key-gen",
                "--verification-key-file",
                str(vkey),
                *extended_args,
                "--signing-key-file",
                str(skey),
            ]
        )

        helpers._check_outfiles(vkey, skey)
        return structs.KeyPair(vkey, skey)

    def get_payment_vkey_hash(
        self,
        payment_vkey_file: itp.FileType | None = None,
        payment_vkey: str | None = None,
    ) -> str:
        """Return the hash of an address key.

        Args:
            payment_vkey_file: A path to payment vkey file (optional).
            payment_vkey: A payment vkey, (Bech32, optional).

        Returns:
            str: A generated hash.
        """
        if payment_vkey:
            cli_args = ["--payment-verification-key", payment_vkey]
        elif payment_vkey_file:
            cli_args = ["--payment-verification-key-file", str(payment_vkey_file)]
        else:
            msg = "Either `payment_vkey` or `payment_vkey_file` is needed."
            raise AssertionError(msg)

        return (
            self._clusterlib_obj.cli(["address", "key-hash", *cli_args])
            .stdout.rstrip()
            .decode("ascii")
        )

    def get_address_info(
        self,
        address: str,
    ) -> structs.AddressInfo:
        """Get information about an address.

        Args:
            address: A Cardano address.

        Returns:
            structs.AddressInfo: A data container containing address info.
        """
        addr_dict: dict[str, str] = json.loads(
            self._clusterlib_obj.cli(["address", "info", "--address", str(address)])
            .stdout.rstrip()
            .decode("utf-8")
        )
        return structs.AddressInfo(**addr_dict)

    def gen_payment_addr_and_keys(
        self,
        name: str,
        stake_vkey_file: itp.FileType | None = None,
        stake_script_file: itp.FileType | None = None,
        destination_dir: itp.FileType = ".",
    ) -> structs.AddressRecord:
        """Generate payment address and key pair.

        Args:
            name: A name of the address and key pair.
            stake_vkey_file: A path to corresponding stake vkey file (optional).
            stake_script_file: A path to corresponding payment script file (optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.AddressRecord: A data container containing the address and
                key pair / script file.
        """
        key_pair = self.gen_payment_key_pair(key_name=name, destination_dir=destination_dir)
        addr = self.gen_payment_addr(
            addr_name=name,
            payment_vkey_file=key_pair.vkey_file,
            stake_vkey_file=stake_vkey_file,
            stake_script_file=stake_script_file,
            destination_dir=destination_dir,
        )

        return structs.AddressRecord(
            address=addr, vkey_file=key_pair.vkey_file, skey_file=key_pair.skey_file
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: clusterlib_obj={id(self._clusterlib_obj)}>"
